/* automatically generated by rust-bindgen 0.59.2 */

pub const FIONREAD: u32 = 18047;
pub const FIONBIO: u32 = 26238;
pub const FIOCLEX: u32 = 26113;
pub const FIONCLEX: u32 = 26114;
pub const FIOASYNC: u32 = 26237;
pub const FIOQSIZE: u32 = 26239;
pub const TCXONC: u32 = 21510;
pub const TCFLSH: u32 = 21511;
pub const TIOCSCTTY: u32 = 21632;
pub const TIOCSPGRP: u32 = 2147775606;
pub const TIOCOUTQ: u32 = 29810;
pub const TIOCSTI: u32 = 21618;
pub const TIOCSWINSZ: u32 = 2148037735;
pub const TIOCMGET: u32 = 29725;
pub const TIOCMBIS: u32 = 29723;
pub const TIOCMBIC: u32 = 29724;
pub const TIOCMSET: u32 = 29722;
pub const TIOCSSOFTCAR: u32 = 21634;
pub const TIOCLINUX: u32 = 21635;
pub const TIOCCONS: u32 = 2147775608;
pub const TIOCSSERIAL: u32 = 21637;
pub const TIOCPKT: u32 = 21616;
pub const TIOCNOTTY: u32 = 21617;
pub const TIOCSETD: u32 = 29697;
pub const TIOCSBRK: u32 = 21543;
pub const TIOCCBRK: u32 = 21544;
pub const TIOCSPTLCK: u32 = 2147767345;
pub const TIOCSIG: u32 = 2147767350;
pub const TIOCVHANGUP: u32 = 21559;
pub const TIOCSERCONFIG: u32 = 21640;
pub const TIOCSERGWILD: u32 = 21641;
pub const TIOCSERSWILD: u32 = 21642;
pub const TIOCSLCKTRMIOS: u32 = 21644;
pub const TIOCSERGSTRUCT: u32 = 21645;
pub const TIOCSERGETLSR: u32 = 21646;
pub const TIOCSERGETMULTI: u32 = 21647;
pub const TIOCSERSETMULTI: u32 = 21648;
pub const TIOCMIWAIT: u32 = 21649;
pub const TCGETS: u32 = 21517;
pub const TCGETA: u32 = 21505;
pub const TCSBRK: u32 = 21509;
pub const TCSBRKP: u32 = 21638;
pub const TCSETA: u32 = 21506;
pub const TCSETAF: u32 = 21508;
pub const TCSETAW: u32 = 21507;
pub const TIOCEXCL: u32 = 29709;
pub const TIOCNXCL: u32 = 29710;
pub const TIOCGDEV: u32 = 1074025522;
pub const TIOCGEXCL: u32 = 1074025536;
pub const TIOCGICOUNT: u32 = 21650;
pub const TIOCGLCKTRMIOS: u32 = 21643;
pub const TIOCGPGRP: u32 = 1074033783;
pub const TIOCGPKT: u32 = 1074025528;
pub const TIOCGPTLCK: u32 = 1074025529;
pub const TIOCGPTN: u32 = 1074025520;
pub const TIOCGPTPEER: u32 = 536892481;
pub const TIOCGSERIAL: u32 = 21636;
pub const TIOCGSID: u32 = 29718;
pub const TIOCGSOFTCAR: u32 = 21633;
pub const TIOCGWINSZ: u32 = 1074295912;
pub const TCGETS2: u32 = 1076909098;
pub const TCSETS: u32 = 21518;
pub const TCSETS2: u32 = 2150650923;
pub const TCSETSF: u32 = 21520;
pub const TCSETSF2: u32 = 2150650925;
pub const TCSETSW: u32 = 21519;
pub const TCSETSW2: u32 = 2150650924;
pub const TIOCGETD: u32 = 29696;
pub const TIOCGETP: u32 = 29704;
pub const TIOCGLTC: u32 = 29812;
pub const MTIOCGET: u32 = 1075604738;
pub const BLKSSZGET: u32 = 536875624;
pub const BLKPBSZGET: u32 = 536875643;
pub const BLKROSET: u32 = 536875613;
pub const BLKROGET: u32 = 536875614;
pub const BLKRRPART: u32 = 536875615;
pub const BLKGETSIZE: u32 = 536875616;
pub const BLKFLSBUF: u32 = 536875617;
pub const BLKRASET: u32 = 536875618;
pub const BLKRAGET: u32 = 536875619;
pub const BLKFRASET: u32 = 536875620;
pub const BLKFRAGET: u32 = 536875621;
pub const BLKSECTSET: u32 = 536875622;
pub const BLKSECTGET: u32 = 536875623;
pub const BLKPG: u32 = 536875625;
pub const BLKBSZGET: u32 = 1074008688;
pub const BLKBSZSET: u32 = 2147750513;
pub const BLKGETSIZE64: u32 = 1074008690;
pub const BLKTRACESETUP: u32 = 3225948787;
pub const BLKTRACESTART: u32 = 536875636;
pub const BLKTRACESTOP: u32 = 536875637;
pub const BLKTRACETEARDOWN: u32 = 536875638;
pub const BLKDISCARD: u32 = 536875639;
pub const BLKIOMIN: u32 = 536875640;
pub const BLKIOOPT: u32 = 536875641;
pub const BLKALIGNOFF: u32 = 536875642;
pub const BLKDISCARDZEROES: u32 = 536875644;
pub const BLKSECDISCARD: u32 = 536875645;
pub const BLKROTATIONAL: u32 = 536875646;
pub const BLKZEROOUT: u32 = 536875647;
pub const UFFDIO_REGISTER: u32 = 3223366144;
pub const UFFDIO_UNREGISTER: u32 = 1074833921;
pub const UFFDIO_WAKE: u32 = 1074833922;
pub const UFFDIO_COPY: u32 = 3223890435;
pub const UFFDIO_ZEROPAGE: u32 = 3223366148;
pub const UFFDIO_WRITEPROTECT: u32 = 3222841862;
pub const UFFDIO_API: u32 = 3222841919;
pub const NS_GET_USERNS: u32 = 536917761;
pub const NS_GET_PARENT: u32 = 536917762;
pub const NS_GET_NSTYPE: u32 = 536917763;
pub const KDGETLED: u32 = 19249;
pub const KDSETLED: u32 = 19250;
pub const KDGKBLED: u32 = 19300;
pub const KDSKBLED: u32 = 19301;
pub const KDGKBTYPE: u32 = 19251;
pub const KDADDIO: u32 = 19252;
pub const KDDELIO: u32 = 19253;
pub const KDENABIO: u32 = 19254;
pub const KDDISABIO: u32 = 19255;
pub const KDSETMODE: u32 = 19258;
pub const KDGETMODE: u32 = 19259;
pub const KDMKTONE: u32 = 19248;
pub const KIOCSOUND: u32 = 19247;
pub const GIO_CMAP: u32 = 19312;
pub const PIO_CMAP: u32 = 19313;
pub const GIO_FONT: u32 = 19296;
pub const GIO_FONTX: u32 = 19307;
pub const PIO_FONT: u32 = 19297;
pub const PIO_FONTX: u32 = 19308;
pub const PIO_FONTRESET: u32 = 19309;
pub const GIO_SCRNMAP: u32 = 19264;
pub const GIO_UNISCRNMAP: u32 = 19305;
pub const PIO_SCRNMAP: u32 = 19265;
pub const PIO_UNISCRNMAP: u32 = 19306;
pub const GIO_UNIMAP: u32 = 19302;
pub const PIO_UNIMAP: u32 = 19303;
pub const PIO_UNIMAPCLR: u32 = 19304;
pub const KDGKBMODE: u32 = 19268;
pub const KDSKBMODE: u32 = 19269;
pub const KDGKBMETA: u32 = 19298;
pub const KDSKBMETA: u32 = 19299;
pub const KDGKBENT: u32 = 19270;
pub const KDSKBENT: u32 = 19271;
pub const KDGKBSENT: u32 = 19272;
pub const KDSKBSENT: u32 = 19273;
pub const KDGKBDIACR: u32 = 19274;
pub const KDGETKEYCODE: u32 = 19276;
pub const KDSETKEYCODE: u32 = 19277;
pub const KDSIGACCEPT: u32 = 19278;
pub const VT_OPENQRY: u32 = 22016;
pub const VT_GETMODE: u32 = 22017;
pub const VT_SETMODE: u32 = 22018;
pub const VT_GETSTATE: u32 = 22019;
pub const VT_RELDISP: u32 = 22021;
pub const VT_ACTIVATE: u32 = 22022;
pub const VT_WAITACTIVE: u32 = 22023;
pub const VT_DISALLOCATE: u32 = 22024;
pub const VT_RESIZE: u32 = 22025;
pub const VT_RESIZEX: u32 = 22026;
pub const FIOSETOWN: u32 = 2147772028;
pub const FIOGETOWN: u32 = 1074030203;
pub const SIOCATMARK: u32 = 1074033415;
pub const SIOCGSTAMP: u32 = 35078;
pub const TIOCINQ: u32 = 18047;
pub const SIOCADDRT: u32 = 35083;
pub const SIOCDELRT: u32 = 35084;
pub const SIOCGIFNAME: u32 = 35088;
pub const SIOCSIFLINK: u32 = 35089;
pub const SIOCGIFCONF: u32 = 35090;
pub const SIOCGIFFLAGS: u32 = 35091;
pub const SIOCSIFFLAGS: u32 = 35092;
pub const SIOCGIFADDR: u32 = 35093;
pub const SIOCSIFADDR: u32 = 35094;
pub const SIOCGIFDSTADDR: u32 = 35095;
pub const SIOCSIFDSTADDR: u32 = 35096;
pub const SIOCGIFBRDADDR: u32 = 35097;
pub const SIOCSIFBRDADDR: u32 = 35098;
pub const SIOCGIFNETMASK: u32 = 35099;
pub const SIOCSIFNETMASK: u32 = 35100;
pub const SIOCGIFMETRIC: u32 = 35101;
pub const SIOCSIFMETRIC: u32 = 35102;
pub const SIOCGIFMEM: u32 = 35103;
pub const SIOCSIFMEM: u32 = 35104;
pub const SIOCGIFMTU: u32 = 35105;
pub const SIOCSIFMTU: u32 = 35106;
pub const SIOCSIFHWADDR: u32 = 35108;
pub const SIOCGIFENCAP: u32 = 35109;
pub const SIOCSIFENCAP: u32 = 35110;
pub const SIOCGIFHWADDR: u32 = 35111;
pub const SIOCGIFSLAVE: u32 = 35113;
pub const SIOCSIFSLAVE: u32 = 35120;
pub const SIOCADDMULTI: u32 = 35121;
pub const SIOCDELMULTI: u32 = 35122;
pub const SIOCDARP: u32 = 35155;
pub const SIOCGARP: u32 = 35156;
pub const SIOCSARP: u32 = 35157;
pub const SIOCDRARP: u32 = 35168;
pub const SIOCGRARP: u32 = 35169;
pub const SIOCSRARP: u32 = 35170;
pub const SIOCGIFMAP: u32 = 35184;
pub const SIOCSIFMAP: u32 = 35185;
pub const FIBMAP: u32 = 536870913;
pub const FIGETBSZ: u32 = 536870914;
pub const FIFREEZE: u32 = 3221510263;
pub const FITHAW: u32 = 3221510264;
pub const FITRIM: u32 = 3222820985;
pub const FICLONE: u32 = 2147783689;
pub const FICLONERANGE: u32 = 2149618701;
pub const FIDEDUPERANGE: u32 = 3222836278;
pub const FS_IOC_GETFLAGS: u32 = 1074030081;
pub const FS_IOC_SETFLAGS: u32 = 2147771906;
pub const FS_IOC_GETVERSION: u32 = 1074034177;
pub const FS_IOC_SETVERSION: u32 = 2147776002;
pub const FS_IOC_FIEMAP: u32 = 3223348747;
pub const FS_IOC32_GETFLAGS: u32 = 1074030081;
pub const FS_IOC32_SETFLAGS: u32 = 2147771906;
pub const FS_IOC32_GETVERSION: u32 = 1074034177;
pub const FS_IOC32_SETVERSION: u32 = 2147776002;
pub const FS_IOC_FSGETXATTR: u32 = 1075599391;
pub const FS_IOC_FSSETXATTR: u32 = 2149341216;
pub const FS_IOC_GETFSLABEL: u32 = 1090556977;
pub const FS_IOC_SETFSLABEL: u32 = 2164298802;
pub const VIDIOC_SUBDEV_QUERYSTD: u32 = 1074288191;
pub const AUTOFS_DEV_IOCTL_CLOSEMOUNT: u32 = 3222836085;
pub const LIRC_SET_SEND_CARRIER: u32 = 2147772691;
pub const AUTOFS_IOC_PROTOSUBVER: u32 = 1074041703;
pub const PTP_SYS_OFFSET_PRECISE: u32 = 3225435400;
pub const FSI_SCOM_WRITE: u32 = 3223352066;
pub const ATM_GETCIRANGE: u32 = 2148295050;
pub const DMA_BUF_SET_NAME_B: u32 = 2148033025;
pub const RIO_CM_EP_GET_LIST_SIZE: u32 = 3221512961;
pub const TUNSETPERSIST: u32 = 2147767499;
pub const FS_IOC_GET_ENCRYPTION_POLICY: u32 = 2148296213;
pub const CEC_RECEIVE: u32 = 3224920326;
pub const MGSL_IOCGPARAMS: u32 = 1075866881;
pub const ENI_SETMULT: u32 = 2148295015;
pub const RIO_GET_EVENT_MASK: u32 = 1074031886;
pub const LIRC_GET_MAX_TIMEOUT: u32 = 1074030857;
pub const KVM_SET_BOOT_CPU_ID: u32 = 536915576;
pub const KVM_SET_MP_STATE: u32 = 2147790489;
pub const KVM_GET_MP_STATE: u32 = 1074048664;
pub const KVM_SET_MEMORY_REGION: u32 = 2149101120;
pub const KVM_GET_API_VERSION: u32 = 536915456;
pub const KVM_S390_GET_SKEYS: u32 = 2151722674;
pub const KVM_SET_ONE_REG: u32 = 2148576940;
pub const KVM_IRQ_LINE_STATUS: u32 = 3221794407;
pub const KVM_SET_USER_MEMORY_REGION: u32 = 2149625414;
pub const KVM_S390_INTERRUPT: u32 = 2148576916;
pub const KVM_TPR_ACCESS_REPORTING: u32 = 3223891602;
pub const KVM_GET_FPU: u32 = 1073786508;
pub const KVM_ASSIGN_SET_MSIX_ENTRY: u32 = 2148576884;
pub const KVM_MEMORY_ENCRYPT_REG_REGION: u32 = 1074835131;
pub const KVM_S390_UCAS_MAP: u32 = 2149101136;
pub const KVM_ASSIGN_SET_INTX_MASK: u32 = 2151722660;
pub const KVM_MEMORY_ENCRYPT_OP: u32 = 3221532346;
pub const KVM_CREATE_PIT2: u32 = 2151722615;
pub const KVM_S390_VCPU_FAULT: u32 = 2147790418;
pub const KVM_REGISTER_COALESCED_MMIO: u32 = 2148576871;
pub const KVM_SET_VAPIC_ADDR: u32 = 2148052627;
pub const KVM_CREATE_DEVICE: u32 = 3222056672;
pub const KVM_GET_IRQCHIP: u32 = 3255348834;
pub const KVM_TRANSLATE: u32 = 3222843013;
pub const KVM_S390_SET_INITIAL_PSW: u32 = 2148576918;
pub const KVM_X86_SETUP_MCE: u32 = 2148052636;
pub const KVM_DEASSIGN_PCI_DEVICE: u32 = 2151722610;
pub const KVM_XEN_HVM_SET_ATTR: u32 = 2152246985;
pub const KVM_GET_CLOCK: u32 = 1076932220;
pub const KVM_SET_FPU: u32 = 2147528333;
pub const KVM_KVMCLOCK_CTRL: u32 = 536915629;
pub const KVM_PPC_GET_PVINFO: u32 = 2155916961;
pub const KVM_X86_GET_MCE_CAP_SUPPORTED: u32 = 1074310813;
pub const KVM_S390_GET_IRQ_STATE: u32 = 2149625526;
pub const KVM_GET_DEVICE_ATTR: u32 = 2149101282;
pub const KVM_S390_INITIAL_RESET: u32 = 536915607;
pub const KVM_SET_REGS: u32 = 2165878402;
pub const KVM_GET_TSC_KHZ: u32 = 536915619;
pub const KVM_CHECK_EXTENSION: u32 = 536915459;
pub const KVM_PPC_RESIZE_HPT_COMMIT: u32 = 1074835118;
pub const KVM_GET_DIRTY_LOG: u32 = 2148576834;
pub const KVM_S390_SET_IRQ_STATE: u32 = 2149625525;
pub const KVM_S390_NORMAL_RESET: u32 = 536915651;
pub const KVM_SET_SIGNAL_MASK: u32 = 2147790475;
pub const KVM_XEN_VCPU_SET_ATTR: u32 = 2152246987;
pub const KVM_SET_GSI_ROUTING: u32 = 2148052586;
pub const KVM_CREATE_VCPU: u32 = 536915521;
pub const KVM_S390_UCAS_UNMAP: u32 = 2149101137;
pub const KVM_S390_STORE_STATUS: u32 = 2147790485;
pub const KVM_UNREGISTER_COALESCED_MMIO: u32 = 2148576872;
pub const KVM_ASSIGN_DEV_IRQ: u32 = 2151722608;
pub const KVM_S390_PV_COMMAND: u32 = 3223367365;
pub const KVM_S390_ENABLE_SIE: u32 = 536915462;
pub const KVM_CLEAR_DIRTY_LOG: u32 = 3222843072;
pub const KVM_GET_ONE_REG: u32 = 2148576939;
pub const KVM_ARM_SET_DEVICE_ADDR: u32 = 2148576939;
pub const KVM_DIRTY_TLB: u32 = 2148576938;
pub const KVM_NMI: u32 = 536915610;
pub const KVM_HYPERV_EVENTFD: u32 = 2149101245;
pub const KVM_S390_MEM_OP: u32 = 2151722673;
pub const KVM_PPC_SVM_OFF: u32 = 536915635;
pub const KVM_PPC_RESIZE_HPT_PREPARE: u32 = 1074835117;
pub const KVM_INTERRUPT: u32 = 2147790470;
pub const KVM_XEN_HVM_GET_ATTR: u32 = 3225988808;
pub const KVM_GET_SREGS: u32 = 1073786499;
pub const KVM_SET_IDENTITY_MAP_ADDR: u32 = 2148052552;
pub const KVM_GET_REGS: u32 = 1092136577;
pub const KVM_S390_GET_CMMA_BITS: u32 = 3223367352;
pub const KVM_CREATE_IRQCHIP: u32 = 536915552;
pub const KVM_ASSIGN_PCI_DEVICE: u32 = 1077980777;
pub const KVM_SET_NR_MMU_PAGES: u32 = 536915524;
pub const KVM_IOEVENTFD: u32 = 2151722617;
pub const KVM_RUN: u32 = 536915584;
pub const KVM_S390_SET_SKEYS: u32 = 2151722675;
pub const KVM_SET_DEVICE_ATTR: u32 = 2149101281;
pub const KVM_DEASSIGN_DEV_IRQ: u32 = 2151722613;
pub const KVM_SMI: u32 = 536915639;
pub const KVM_SET_TSS_ADDR: u32 = 536915527;
pub const KVM_GET_NR_MMU_PAGES: u32 = 536915525;
pub const KVM_XEN_VCPU_GET_ATTR: u32 = 3225988810;
pub const KVM_REINJECT_CONTROL: u32 = 536915569;
pub const KVM_HAS_DEVICE_ATTR: u32 = 2149101283;
pub const KVM_IRQFD: u32 = 2149625462;
pub const KVM_S390_SET_CMMA_BITS: u32 = 2149625529;
pub const KVM_SET_IRQCHIP: u32 = 1107865187;
pub const KVM_SIGNAL_MSI: u32 = 2149625509;
pub const KVM_GET_VCPU_MMAP_SIZE: u32 = 536915460;
pub const KVM_CREATE_VM: u32 = 536915457;
pub const KVM_S390_CLEAR_RESET: u32 = 536915652;
pub const KVM_RESET_DIRTY_RINGS: u32 = 536915655;
pub const KVM_ASSIGN_SET_MSIX_NR: u32 = 2148052595;
pub const KVM_CREATE_PIT: u32 = 536915556;
pub const KVM_SET_GUEST_DEBUG: u32 = 2148052635;
pub const KVM_GET_REG_LIST: u32 = 3221794480;
pub const KVM_ARM_VCPU_FINALIZE: u32 = 2147790530;
pub const KVM_SET_CLOCK: u32 = 2150674043;
pub const KVM_PPC_ALLOCATE_HTAB: u32 = 3221532327;
pub const KVM_PPC_GET_SMMU_INFO: u32 = 1112583846;
pub const KVM_SET_TSC_KHZ: u32 = 536915618;
pub const KVM_S390_IRQ: u32 = 2152246964;
pub const KVM_ENABLE_CAP: u32 = 2154344099;
pub const KVM_IRQ_LINE: u32 = 2148052577;
pub const KVM_SET_SREGS: u32 = 2147528324;
pub const KVM_MEMORY_ENCRYPT_UNREG_REGION: u32 = 1074835132;
pub const USBDEVFS_CLAIMINTERFACE: u32 = 1074025743;
pub const CHIOMOVE: u32 = 2148819713;
pub const SONYPI_IOCGBATFLAGS: u32 = 1073837575;
pub const BTRFS_IOC_SYNC: u32 = 536908808;
pub const VIDIOC_TRY_FMT: u32 = 3234616896;
pub const LIRC_SET_REC_MODE: u32 = 2147772690;
pub const VIDIOC_DQEVENT: u32 = 1082152537;
pub const RPMSG_DESTROY_EPT_IOCTL: u32 = 536917250;
pub const UVCIOC_CTRL_MAP: u32 = 3227022624;
pub const VHOST_SET_BACKEND_FEATURES: u32 = 2148052773;
pub const VHOST_VSOCK_SET_GUEST_CID: u32 = 2148052832;
pub const UI_SET_KEYBIT: u32 = 2147767653;
pub const LIRC_SET_REC_TIMEOUT: u32 = 2147772696;
pub const FS_IOC_GET_ENCRYPTION_KEY_STATUS: u32 = 3229640218;
pub const BTRFS_IOC_TREE_SEARCH_V2: u32 = 3228603409;
pub const VHOST_SET_VRING_BASE: u32 = 2148052754;
pub const RIO_ENABLE_DOORBELL_RANGE: u32 = 2148035849;
pub const VIDIOC_TRY_EXT_CTRLS: u32 = 3222820425;
pub const LIRC_GET_REC_MODE: u32 = 1074030850;
pub const PPGETTIME: u32 = 1074294933;
pub const BTRFS_IOC_RM_DEV: u32 = 2415957003;
pub const ATM_SETBACKEND: u32 = 2147639794;
pub const FSL_HV_IOCTL_PARTITION_START: u32 = 3222318851;
pub const FBIO_WAITEVENT: u32 = 536888968;
pub const SWITCHTEC_IOCTL_PORT_TO_PFF: u32 = 3222034245;
pub const NVME_IOCTL_IO_CMD: u32 = 3225964099;
pub const IPMICTL_RECEIVE_MSG_TRUNC: u32 = 3222825227;
pub const FDTWADDLE: u32 = 536871513;
pub const NVME_IOCTL_SUBMIT_IO: u32 = 2150649410;
pub const NILFS_IOCTL_SYNC: u32 = 1074294410;
pub const VIDIOC_SUBDEV_S_DV_TIMINGS: u32 = 3229898327;
pub const ASPEED_LPC_CTRL_IOCTL_GET_SIZE: u32 = 3222319616;
pub const DM_DEV_STATUS: u32 = 3241737479;
pub const TEE_IOC_CLOSE_SESSION: u32 = 1074045957;
pub const NS_GETPSTAT: u32 = 3222036833;
pub const UI_SET_PROPBIT: u32 = 2147767662;
pub const TUNSETFILTEREBPF: u32 = 1074025697;
pub const RIO_MPORT_MAINT_COMPTAG_SET: u32 = 2147773698;
pub const AUTOFS_DEV_IOCTL_VERSION: u32 = 3222836081;
pub const WDIOC_SETOPTIONS: u32 = 1074026244;
pub const VHOST_SCSI_SET_ENDPOINT: u32 = 2162732864;
pub const MGSL_IOCGTXIDLE: u32 = 536898819;
pub const ATM_ADDLECSADDR: u32 = 2148295054;
pub const FSL_HV_IOCTL_GETPROP: u32 = 3223891719;
pub const FDGETPRM: u32 = 1075577348;
pub const HIDIOCAPPLICATION: u32 = 536889346;
pub const ENI_MEMDUMP: u32 = 2148295008;
pub const PTP_SYS_OFFSET2: u32 = 2202025230;
pub const VIDIOC_SUBDEV_G_DV_TIMINGS: u32 = 3229898328;
pub const DMA_BUF_SET_NAME_A: u32 = 2147770881;
pub const PTP_PIN_GETFUNC: u32 = 3227532550;
pub const PTP_SYS_OFFSET_EXTENDED: u32 = 3300932873;
pub const DFL_FPGA_PORT_UINT_SET_IRQ: u32 = 2148054600;
pub const RTC_EPOCH_READ: u32 = 1074032653;
pub const VIDIOC_SUBDEV_S_SELECTION: u32 = 3225441854;
pub const VIDIOC_QUERY_EXT_CTRL: u32 = 3236451943;
pub const ATM_GETLECSADDR: u32 = 2148295056;
pub const FSL_HV_IOCTL_PARTITION_STOP: u32 = 3221794564;
pub const SONET_GETDIAG: u32 = 1074028820;
pub const ATMMPC_DATA: u32 = 536895961;
pub const IPMICTL_UNREGISTER_FOR_CMD_CHANS: u32 = 1074555165;
pub const HIDIOCGCOLLECTIONINDEX: u32 = 2149074960;
pub const RPMSG_CREATE_EPT_IOCTL: u32 = 2150151425;
pub const GPIOHANDLE_GET_LINE_VALUES_IOCTL: u32 = 3225465864;
pub const UI_DEV_SETUP: u32 = 2153534723;
pub const ISST_IF_IO_CMD: u32 = 2147810818;
pub const RIO_MPORT_MAINT_READ_REMOTE: u32 = 1075342599;
pub const VIDIOC_OMAP3ISP_HIST_CFG: u32 = 3224393412;
pub const BLKGETNRZONES: u32 = 1074008709;
pub const VIDIOC_G_MODULATOR: u32 = 3225703990;
pub const VBG_IOCTL_WRITE_CORE_DUMP: u32 = 3223082515;
pub const USBDEVFS_SETINTERFACE: u32 = 1074287876;
pub const PPPIOCGCHAN: u32 = 1074033719;
pub const EVIOCGVERSION: u32 = 1074021633;
pub const VHOST_NET_SET_BACKEND: u32 = 2148052784;
pub const USBDEVFS_REAPURBNDELAY: u32 = 2147767565;
pub const RNDZAPENTCNT: u32 = 536891908;
pub const VIDIOC_G_PARM: u32 = 3234616853;
pub const TUNGETDEVNETNS: u32 = 536892643;
pub const LIRC_SET_MEASURE_CARRIER_MODE: u32 = 2147772701;
pub const VHOST_SET_VRING_ERR: u32 = 2148052770;
pub const VDUSE_VQ_SETUP: u32 = 2149613844;
pub const AUTOFS_IOC_SETTIMEOUT: u32 = 3221525348;
pub const VIDIOC_S_FREQUENCY: u32 = 2150389305;
pub const F2FS_IOC_SEC_TRIM_FILE: u32 = 2149119252;
pub const FS_IOC_REMOVE_ENCRYPTION_KEY: u32 = 3225445912;
pub const WDIOC_GETPRETIMEOUT: u32 = 1074026249;
pub const USBDEVFS_DROP_PRIVILEGES: u32 = 2147767582;
pub const BTRFS_IOC_SNAP_CREATE_V2: u32 = 2415957015;
pub const VHOST_VSOCK_SET_RUNNING: u32 = 2147790689;
pub const STP_SET_OPTIONS: u32 = 2148017410;
pub const FBIO_RADEON_GET_MIRROR: u32 = 1074020355;
pub const IVTVFB_IOC_DMA_FRAME: u32 = 2148292288;
pub const IPMICTL_SEND_COMMAND: u32 = 1075079437;
pub const VIDIOC_G_ENC_INDEX: u32 = 1209554508;
pub const DFL_FPGA_FME_PORT_PR: u32 = 536917632;
pub const CHIOSVOLTAG: u32 = 2150654738;
pub const ATM_SETESIF: u32 = 2148295053;
pub const FW_CDEV_IOC_SEND_RESPONSE: u32 = 2149065476;
pub const PMU_IOC_GET_MODEL: u32 = 1074020867;
pub const JSIOCGBTNMAP: u32 = 1140877876;
pub const USBDEVFS_HUB_PORTINFO: u32 = 1082152211;
pub const VBG_IOCTL_INTERRUPT_ALL_WAIT_FOR_EVENTS: u32 = 3222820363;
pub const FDCLRPRM: u32 = 536871489;
pub const BTRFS_IOC_SCRUB: u32 = 3288372251;
pub const USBDEVFS_DISCONNECT: u32 = 536892694;
pub const TUNSETVNETBE: u32 = 2147767518;
pub const ATMTCP_REMOVE: u32 = 536895887;
pub const VHOST_VDPA_GET_CONFIG: u32 = 1074311027;
pub const PPPIOCGNPMODE: u32 = 3221779532;
pub const FDGETDRVPRM: u32 = 1079509521;
pub const TUNSETVNETLE: u32 = 2147767516;
pub const PHN_SETREG: u32 = 2148036614;
pub const PPPIOCDETACH: u32 = 2147775548;
pub const MMTIMER_GETRES: u32 = 1074031873;
pub const VIDIOC_SUBDEV_ENUMSTD: u32 = 3225966105;
pub const PPGETFLAGS: u32 = 1074032794;
pub const VDUSE_DEV_GET_FEATURES: u32 = 1074299153;
pub const CAPI_MANUFACTURER_CMD: u32 = 3221766944;
pub const VIDIOC_G_TUNER: u32 = 3226752541;
pub const DM_TABLE_STATUS: u32 = 3241737484;
pub const DM_DEV_ARM_POLL: u32 = 3241737488;
pub const NE_CREATE_VM: u32 = 1074310688;
pub const MEDIA_IOC_ENUM_LINKS: u32 = 3223092226;
pub const F2FS_IOC_PRECACHE_EXTENTS: u32 = 536933647;
pub const DFL_FPGA_PORT_DMA_MAP: u32 = 536917571;
pub const MGSL_IOCGXCTRL: u32 = 536898838;
pub const FW_CDEV_IOC_SEND_REQUEST: u32 = 2150114049;
pub const SONYPI_IOCGBLUE: u32 = 1073837576;
pub const F2FS_IOC_DECOMPRESS_FILE: u32 = 536933655;
pub const I2OHTML: u32 = 3223087369;
pub const VFIO_GET_API_VERSION: u32 = 536886116;
pub const IDT77105_GETSTATZ: u32 = 2148294963;
pub const I2OPARMSET: u32 = 3222825219;
pub const TEE_IOC_CANCEL: u32 = 1074308100;
pub const PTP_SYS_OFFSET_PRECISE2: u32 = 3225435409;
pub const DFL_FPGA_PORT_RESET: u32 = 536917568;
pub const PPPIOCGASYNCMAP: u32 = 1074033752;
pub const EVIOCGKEYCODE_V2: u32 = 1076380932;
pub const DM_DEV_SET_GEOMETRY: u32 = 3241737487;
pub const HIDIOCSUSAGE: u32 = 2149074956;
pub const FW_CDEV_IOC_DEALLOCATE_ISO_RESOURCE_ONCE: u32 = 2149065488;
pub const PTP_EXTTS_REQUEST: u32 = 2148547842;
pub const SWITCHTEC_IOCTL_EVENT_CTL: u32 = 3223869251;
pub const WDIOC_SETPRETIMEOUT: u32 = 3221509896;
pub const VHOST_SCSI_CLEAR_ENDPOINT: u32 = 2162732865;
pub const JSIOCGAXES: u32 = 1073834513;
pub const HIDIOCSFLAG: u32 = 2147764239;
pub const PTP_PEROUT_REQUEST2: u32 = 2151169292;
pub const PPWDATA: u32 = 2147577990;
pub const PTP_CLOCK_GETCAPS: u32 = 1079000321;
pub const FDGETMAXERRS: u32 = 1075053070;
pub const TUNSETQUEUE: u32 = 2147767513;
pub const PTP_ENABLE_PPS: u32 = 2147761412;
pub const SIOCSIFATMTCP: u32 = 536895872;
pub const CEC_ADAP_G_LOG_ADDRS: u32 = 1079795971;
pub const ND_IOCTL_ARS_CAP: u32 = 3223342593;
pub const NBD_SET_BLKSIZE: u32 = 536914689;
pub const NBD_SET_TIMEOUT: u32 = 536914697;
pub const VHOST_SCSI_GET_ABI_VERSION: u32 = 2147790658;
pub const RIO_UNMAP_INBOUND: u32 = 2148035858;
pub const ATM_QUERYLOOP: u32 = 2148294996;
pub const DFL_FPGA_GET_API_VERSION: u32 = 536917504;
pub const USBDEVFS_WAIT_FOR_RESUME: u32 = 536892707;
pub const FBIO_CURSOR: u32 = 3225961992;
pub const RNDCLEARPOOL: u32 = 536891910;
pub const VIDIOC_QUERYSTD: u32 = 1074288191;
pub const DMA_BUF_IOCTL_SYNC: u32 = 2148033024;
pub const SCIF_RECV: u32 = 3222827783;
pub const PTP_PIN_GETFUNC2: u32 = 3227532559;
pub const FW_CDEV_IOC_ALLOCATE: u32 = 3223331586;
pub const CEC_ADAP_G_CAPS: u32 = 3226231040;
pub const VIDIOC_G_FBUF: u32 = 1076647434;
pub const PTP_ENABLE_PPS2: u32 = 2147761421;
pub const PCITEST_CLEAR_IRQ: u32 = 536891408;
pub const IPMICTL_SET_GETS_EVENTS_CMD: u32 = 1074030864;
pub const BTRFS_IOC_DEVICES_READY: u32 = 1342215207;
pub const JSIOCGAXMAP: u32 = 1077963314;
pub const FW_CDEV_IOC_GET_CYCLE_TIMER: u32 = 1074799372;
pub const FW_CDEV_IOC_SET_ISO_CHANNELS: u32 = 2148541207;
pub const RTC_WIE_OFF: u32 = 536899600;
pub const PPGETMODE: u32 = 1074032792;
pub const VIDIOC_DBG_G_REGISTER: u32 = 3224917584;
pub const PTP_SYS_OFFSET: u32 = 2202025221;
pub const BTRFS_IOC_SPACE_INFO: u32 = 3222311956;
pub const VIDIOC_SUBDEV_ENUM_FRAME_SIZE: u32 = 3225441866;
pub const ND_IOCTL_VENDOR: u32 = 3221769737;
pub const SIOCSNETADDR: u32 = 2147781088;
pub const SCIF_VREADFROM: u32 = 3223876364;
pub const BTRFS_IOC_TRANS_START: u32 = 536908806;
pub const INOTIFY_IOC_SETNEXTWD: u32 = 2147764480;
pub const SNAPSHOT_GET_IMAGE_SIZE: u32 = 1074279182;
pub const TUNDETACHFILTER: u32 = 2148029654;
pub const ND_IOCTL_CLEAR_ERROR: u32 = 3223342596;
pub const IOC_PR_CLEAR: u32 = 2148561101;
pub const SCIF_READFROM: u32 = 3223876362;
pub const PPPIOCGDEBUG: u32 = 1074033729;
pub const BLKGETZONESZ: u32 = 1074008708;
pub const HIDIOCGUSAGES: u32 = 3491514387;
pub const SONYPI_IOCGTEMP: u32 = 1073837580;
pub const UI_SET_MSCBIT: u32 = 2147767656;
pub const APM_IOC_SUSPEND: u32 = 536887554;
pub const BTRFS_IOC_TREE_SEARCH: u32 = 3489698833;
pub const RTC_PLL_GET: u32 = 1075605521;
pub const RIO_CM_EP_GET_LIST: u32 = 3221512962;
pub const USBDEVFS_DISCSIGNAL: u32 = 1074287886;
pub const OSIOCSNETADDR: u32 = 2147781088;
pub const LIRC_GET_MIN_TIMEOUT: u32 = 1074030856;
pub const SWITCHTEC_IOCTL_EVENT_SUMMARY_LEGACY: u32 = 1100502850;
pub const DM_TARGET_MSG: u32 = 3241737486;
pub const SONYPI_IOCGBAT1REM: u32 = 1073903107;
pub const EVIOCSFF: u32 = 2150385024;
pub const TUNSETGROUP: u32 = 2147767502;
pub const EVIOCGKEYCODE: u32 = 1074283780;
pub const KCOV_REMOTE_ENABLE: u32 = 2149081958;
pub const ND_IOCTL_GET_CONFIG_SIZE: u32 = 3222031876;
pub const FDEJECT: u32 = 536871514;
pub const TUNSETOFFLOAD: u32 = 2147767504;
pub const PPPIOCCONNECT: u32 = 2147775546;
pub const ATM_ADDADDR: u32 = 2148295048;
pub const VDUSE_DEV_INJECT_CONFIG_IRQ: u32 = 536903955;
pub const AUTOFS_DEV_IOCTL_ASKUMOUNT: u32 = 3222836093;
pub const VHOST_VDPA_GET_STATUS: u32 = 1073852273;
pub const CCISS_PASSTHRU: u32 = 3226747403;
pub const MGSL_IOCCLRMODCOUNT: u32 = 536898831;
pub const TEE_IOC_SUPPL_SEND: u32 = 1074832391;
pub const ATMARPD_CTRL: u32 = 536895969;
pub const UI_ABS_SETUP: u32 = 2149340420;
pub const UI_DEV_DESTROY: u32 = 536892674;
pub const BTRFS_IOC_QUOTA_CTL: u32 = 3222311976;
pub const RTC_AIE_ON: u32 = 536899585;
pub const AUTOFS_IOC_EXPIRE: u32 = 1091343205;
pub const PPPIOCSDEBUG: u32 = 2147775552;
pub const GPIO_V2_LINE_SET_VALUES_IOCTL: u32 = 3222320143;
pub const PPPIOCSMRU: u32 = 2147775570;
pub const CCISS_DEREGDISK: u32 = 536887820;
pub const UI_DEV_CREATE: u32 = 536892673;
pub const FUSE_DEV_IOC_CLONE: u32 = 1074062592;
pub const BTRFS_IOC_START_SYNC: u32 = 1074304024;
pub const NILFS_IOCTL_DELETE_CHECKPOINT: u32 = 2148036225;
pub const SNAPSHOT_AVAIL_SWAP_SIZE: u32 = 1074279187;
pub const DM_TABLE_CLEAR: u32 = 3241737482;
pub const CCISS_GETINTINFO: u32 = 1074283010;
pub const PPPIOCSASYNCMAP: u32 = 2147775575;
pub const I2OEVTGET: u32 = 1080584459;
pub const NVME_IOCTL_RESET: u32 = 536890948;
pub const PPYIELD: u32 = 536899725;
pub const NVME_IOCTL_IO64_CMD: u32 = 3226488392;
pub const TUNSETCARRIER: u32 = 2147767522;
pub const DM_DEV_WAIT: u32 = 3241737480;
pub const RTC_WIE_ON: u32 = 536899599;
pub const MEDIA_IOC_DEVICE_INFO: u32 = 3238034432;
pub const RIO_CM_CHAN_CREATE: u32 = 3221381891;
pub const MGSL_IOCSPARAMS: u32 = 2149608704;
pub const RTC_SET_TIME: u32 = 2149871626;
pub const VHOST_RESET_OWNER: u32 = 536915714;
pub const IOC_OPAL_PSID_REVERT_TPR: u32 = 2164814056;
pub const AUTOFS_DEV_IOCTL_OPENMOUNT: u32 = 3222836084;
pub const UDF_GETEABLOCK: u32 = 1074031681;
pub const VFIO_IOMMU_MAP_DMA: u32 = 536886129;
pub const VIDIOC_SUBSCRIBE_EVENT: u32 = 2149602906;
pub const HIDIOCGFLAG: u32 = 1074022414;
pub const HIDIOCGUCODE: u32 = 3222816781;
pub const VIDIOC_OMAP3ISP_AF_CFG: u32 = 3226228421;
pub const DM_REMOVE_ALL: u32 = 3241737473;
pub const ASPEED_LPC_CTRL_IOCTL_MAP: u32 = 2148577793;
pub const CCISS_GETFIRMVER: u32 = 1074020872;
pub const ND_IOCTL_ARS_START: u32 = 3223342594;
pub const PPPIOCSMRRU: u32 = 2147775547;
pub const CEC_ADAP_S_LOG_ADDRS: u32 = 3227279620;
pub const RPROC_GET_SHUTDOWN_ON_RELEASE: u32 = 1074050818;
pub const DMA_HEAP_IOCTL_ALLOC: u32 = 3222816768;
pub const PPSETTIME: u32 = 2148036758;
pub const RTC_ALM_READ: u32 = 1076129800;
pub const VDUSE_SET_API_VERSION: u32 = 2148040961;
pub const RIO_MPORT_MAINT_WRITE_REMOTE: u32 = 2149084424;
pub const VIDIOC_SUBDEV_S_CROP: u32 = 3224917564;
pub const USBDEVFS_CONNECT: u32 = 536892695;
pub const SYNC_IOC_FILE_INFO: u32 = 3224911364;
pub const ATMARP_MKIP: u32 = 536895970;
pub const VFIO_IOMMU_SPAPR_TCE_GET_INFO: u32 = 536886128;
pub const CCISS_GETHEARTBEAT: u32 = 1074020870;
pub const ATM_RSTADDR: u32 = 2148295047;
pub const NBD_SET_SIZE: u32 = 536914690;
pub const UDF_GETVOLIDENT: u32 = 1074031682;
pub const GPIO_V2_LINE_GET_VALUES_IOCTL: u32 = 3222320142;
pub const MGSL_IOCSTXIDLE: u32 = 536898818;
pub const FSL_HV_IOCTL_SETPROP: u32 = 3223891720;
pub const BTRFS_IOC_GET_DEV_STATS: u32 = 3288896564;
pub const PPRSTATUS: u32 = 1073836161;
pub const MGSL_IOCTXENABLE: u32 = 536898820;
pub const UDF_GETEASIZE: u32 = 1074031680;
pub const NVME_IOCTL_ADMIN64_CMD: u32 = 3226488391;
pub const VHOST_SET_OWNER: u32 = 536915713;
pub const RIO_ALLOC_DMA: u32 = 3222826259;
pub const F2FS_IOC_ABORT_VOLATILE_WRITE: u32 = 536933637;
pub const RIO_CM_CHAN_ACCEPT: u32 = 3221775111;
pub const I2OHRTGET: u32 = 3222038785;
pub const ATM_SETCIRANGE: u32 = 2148295051;
pub const HPET_IE_ON: u32 = 536897537;
pub const PERF_EVENT_IOC_ID: u32 = 1074013191;
pub const TUNSETSNDBUF: u32 = 2147767508;
pub const PTP_PIN_SETFUNC: u32 = 2153790727;
pub const PPPIOCDISCONN: u32 = 536900665;
pub const VIDIOC_QUERYCTRL: u32 = 3225703972;
pub const PPEXCL: u32 = 536899727;
pub const PCITEST_MSI: u32 = 2147766275;
pub const FDWERRORCLR: u32 = 536871510;
pub const AUTOFS_IOC_FAIL: u32 = 536908641;
pub const USBDEVFS_IOCTL: u32 = 3222033682;
pub const VIDIOC_S_STD: u32 = 2148029976;
pub const F2FS_IOC_RESIZE_FS: u32 = 2148070672;
pub const SONET_SETDIAG: u32 = 3221512466;
pub const BTRFS_IOC_DEFRAG: u32 = 2415956994;
pub const CCISS_GETDRIVVER: u32 = 1074020873;
pub const IPMICTL_GET_TIMING_PARMS_CMD: u32 = 1074293015;
pub const HPET_IRQFREQ: u32 = 2147772422;
pub const ATM_GETESI: u32 = 2148295045;
pub const CCISS_GETLUNINFO: u32 = 1074545169;
pub const AUTOFS_DEV_IOCTL_ISMOUNTPOINT: u32 = 3222836094;
pub const TEE_IOC_SHM_ALLOC: u32 = 3222316033;
pub const PERF_EVENT_IOC_SET_BPF: u32 = 2147755016;
pub const UDMABUF_CREATE_LIST: u32 = 2148037955;
pub const VHOST_SET_LOG_BASE: u32 = 2148052740;
pub const ZATM_GETPOOL: u32 = 2148295009;
pub const BR2684_SETFILT: u32 = 2149343632;
pub const RNDGETPOOL: u32 = 1074287106;
pub const PPS_GETPARAMS: u32 = 1074032801;
pub const IOC_PR_RESERVE: u32 = 2148561097;
pub const VIDIOC_TRY_DECODER_CMD: u32 = 3225966177;
pub const RIO_CM_CHAN_CLOSE: u32 = 2147640068;
pub const VIDIOC_DV_TIMINGS_CAP: u32 = 3230684772;
pub const IOCTL_MEI_CONNECT_CLIENT_VTAG: u32 = 3222554628;
pub const PMU_IOC_GET_BACKLIGHT: u32 = 1074020865;
pub const USBDEVFS_GET_CAPABILITIES: u32 = 1074025754;
pub const SCIF_WRITETO: u32 = 3223876363;
pub const UDF_RELOCATE_BLOCKS: u32 = 3221515331;
pub const FSL_HV_IOCTL_PARTITION_RESTART: u32 = 3221794561;
pub const CCISS_REGNEWD: u32 = 536887822;
pub const FAT_IOCTL_SET_ATTRIBUTES: u32 = 2147774993;
pub const VIDIOC_CREATE_BUFS: u32 = 3237500508;
pub const CAPI_GET_VERSION: u32 = 3222291207;
pub const SWITCHTEC_IOCTL_EVENT_SUMMARY: u32 = 1155028802;
pub const VFIO_EEH_PE_OP: u32 = 536886137;
pub const FW_CDEV_IOC_CREATE_ISO_CONTEXT: u32 = 3223331592;
pub const F2FS_IOC_RELEASE_COMPRESS_BLOCKS: u32 = 1074328850;
pub const NBD_SET_SIZE_BLOCKS: u32 = 536914695;
pub const IPMI_BMC_IOCTL_SET_SMS_ATN: u32 = 536916224;
pub const ASPEED_P2A_CTRL_IOCTL_GET_MEMORY_CONFIG: u32 = 3222319873;
pub const VIDIOC_S_AUDOUT: u32 = 2150913586;
pub const VIDIOC_S_FMT: u32 = 3234616837;
pub const PPPIOCATTACH: u32 = 2147775549;
pub const VHOST_GET_VRING_BUSYLOOP_TIMEOUT: u32 = 2148052772;
pub const FS_IOC_MEASURE_VERITY: u32 = 3221513862;
pub const CCISS_BIG_PASSTHRU: u32 = 3227009554;
pub const IPMICTL_SET_MY_LUN_CMD: u32 = 1074030867;
pub const PCITEST_LEGACY_IRQ: u32 = 536891394;
pub const USBDEVFS_SUBMITURB: u32 = 1076647178;
pub const AUTOFS_IOC_READY: u32 = 536908640;
pub const BTRFS_IOC_SEND: u32 = 2152240166;
pub const VIDIOC_G_EXT_CTRLS: u32 = 3222820423;
pub const JSIOCSBTNMAP: u32 = 2214619699;
pub const PPPIOCSFLAGS: u32 = 2147775577;
pub const NVRAM_INIT: u32 = 536899648;
pub const RFKILL_IOCTL_NOINPUT: u32 = 536891905;
pub const BTRFS_IOC_BALANCE: u32 = 2415957004;
pub const FS_IOC_GETFSMAP: u32 = 3233830971;
pub const IPMICTL_GET_MY_CHANNEL_LUN_CMD: u32 = 1074030875;
pub const STP_POLICY_ID_GET: u32 = 1074799873;
pub const PPSETFLAGS: u32 = 2147774619;
pub const CEC_ADAP_S_PHYS_ADDR: u32 = 2147639554;
pub const ATMTCP_CREATE: u32 = 536895886;
pub const IPMI_BMC_IOCTL_FORCE_ABORT: u32 = 536916226;
pub const PPPIOCGXASYNCMAP: u32 = 1075868752;
pub const VHOST_SET_VRING_CALL: u32 = 2148052769;
pub const LIRC_GET_FEATURES: u32 = 1074030848;
pub const GSMIOC_DISABLE_NET: u32 = 536889091;
pub const AUTOFS_IOC_CATATONIC: u32 = 536908642;
pub const NBD_DO_IT: u32 = 536914691;
pub const LIRC_SET_REC_CARRIER_RANGE: u32 = 2147772703;
pub const IPMICTL_GET_MY_CHANNEL_ADDRESS_CMD: u32 = 1074030873;
pub const EVIOCSCLOCKID: u32 = 2147763616;
pub const USBDEVFS_FREE_STREAMS: u32 = 1074287901;
pub const FSI_SCOM_RESET: u32 = 2147775235;
pub const PMU_IOC_GRAB_BACKLIGHT: u32 = 1074020870;
pub const VIDIOC_SUBDEV_S_FMT: u32 = 3227014661;
pub const FDDEFPRM: u32 = 2149319235;
pub const TEE_IOC_INVOKE: u32 = 1074832387;
pub const USBDEVFS_BULK: u32 = 3222295810;
pub const SCIF_VWRITETO: u32 = 3223876365;
pub const SONYPI_IOCSBRT: u32 = 2147579392;
pub const BTRFS_IOC_FILE_EXTENT_SAME: u32 = 3222836278;
pub const RTC_PIE_ON: u32 = 536899589;
pub const BTRFS_IOC_SCAN_DEV: u32 = 2415956996;
pub const PPPIOCXFERUNIT: u32 = 536900686;
pub const WDIOC_GETTIMEOUT: u32 = 1074026247;
pub const BTRFS_IOC_SET_RECEIVED_SUBVOL: u32 = 3234370597;
pub const DFL_FPGA_PORT_ERR_SET_IRQ: u32 = 2148054598;
pub const FBIO_WAITFORVSYNC: u32 = 2147763744;
pub const RTC_PIE_OFF: u32 = 536899590;
pub const EVIOCGRAB: u32 = 2147763600;
pub const PMU_IOC_SET_BACKLIGHT: u32 = 2147762690;
pub const EVIOCGREP: u32 = 1074283779;
pub const PERF_EVENT_IOC_MODIFY_ATTRIBUTES: u32 = 2147755019;
pub const UFFDIO_CONTINUE: u32 = 3223366151;
pub const VDUSE_GET_API_VERSION: u32 = 1074299136;
pub const RTC_RD_TIME: u32 = 1076129801;
pub const FDMSGOFF: u32 = 536871494;
pub const IPMICTL_REGISTER_FOR_CMD_CHANS: u32 = 1074555164;
pub const CAPI_GET_ERRCODE: u32 = 1073890081;
pub const PCITEST_SET_IRQTYPE: u32 = 2147766280;
pub const VIDIOC_SUBDEV_S_EDID: u32 = 3223606825;
pub const MATROXFB_SET_OUTPUT_MODE: u32 = 2147774202;
pub const RIO_DEV_ADD: u32 = 2149608727;
pub const VIDIOC_ENUM_FREQ_BANDS: u32 = 3225441893;
pub const FBIO_RADEON_SET_MIRROR: u32 = 2147762180;
pub const PCITEST_GET_IRQTYPE: u32 = 536891401;
pub const JSIOCGVERSION: u32 = 1074031105;
pub const SONYPI_IOCSBLUE: u32 = 2147579401;
pub const SNAPSHOT_PREF_IMAGE_SIZE: u32 = 536883986;
pub const F2FS_IOC_GET_FEATURES: u32 = 1074066700;
pub const SCIF_REG: u32 = 3223876360;
pub const NILFS_IOCTL_CLEAN_SEGMENTS: u32 = 2155376264;
pub const FW_CDEV_IOC_INITIATE_BUS_RESET: u32 = 2147754757;
pub const RIO_WAIT_FOR_ASYNC: u32 = 2148035862;
pub const VHOST_SET_VRING_NUM: u32 = 2148052752;
pub const AUTOFS_DEV_IOCTL_PROTOVER: u32 = 3222836082;
pub const RIO_FREE_DMA: u32 = 2148035860;
pub const MGSL_IOCRXENABLE: u32 = 536898821;
pub const IOCTL_VM_SOCKETS_GET_LOCAL_CID: u32 = 536872889;
pub const IPMICTL_SET_TIMING_PARMS_CMD: u32 = 1074293014;
pub const PPPIOCGL2TPSTATS: u32 = 1078490166;
pub const PERF_EVENT_IOC_PERIOD: u32 = 2148017156;
pub const PTP_PIN_SETFUNC2: u32 = 2153790736;
pub const CHIOEXCHANGE: u32 = 2149344002;
pub const NILFS_IOCTL_GET_SUINFO: u32 = 1075342980;
pub const CEC_DQEVENT: u32 = 3226493191;
pub const UI_SET_SWBIT: u32 = 2147767661;
pub const VHOST_VDPA_SET_CONFIG: u32 = 2148052852;
pub const TUNSETIFF: u32 = 2147767498;
pub const CHIOPOSITION: u32 = 2148295427;
pub const IPMICTL_SET_MAINTENANCE_MODE_CMD: u32 = 2147772703;
pub const BTRFS_IOC_DEFAULT_SUBVOL: u32 = 2148045843;
pub const RIO_UNMAP_OUTBOUND: u32 = 2150133008;
pub const CAPI_CLR_FLAGS: u32 = 1074021157;
pub const FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE_ONCE: u32 = 2149065487;
pub const MATROXFB_GET_OUTPUT_CONNECTION: u32 = 1074032376;
pub const EVIOCSMASK: u32 = 2148550035;
pub const BTRFS_IOC_FORGET_DEV: u32 = 2415956997;
pub const CXL_MEM_QUERY_COMMANDS: u32 = 1074318849;
pub const CEC_S_MODE: u32 = 2147770633;
pub const MGSL_IOCSIF: u32 = 536898826;
pub const SWITCHTEC_IOCTL_PFF_TO_PORT: u32 = 3222034244;
pub const PPSETMODE: u32 = 2147774592;
pub const VFIO_DEVICE_SET_IRQS: u32 = 536886126;
pub const VIDIOC_PREPARE_BUF: u32 = 3225704029;
pub const CEC_ADAP_G_CONNECTOR_INFO: u32 = 1078223114;
pub const IOC_OPAL_WRITE_SHADOW_MBR: u32 = 2166386922;
pub const VIDIOC_SUBDEV_ENUM_FRAME_INTERVAL: u32 = 3225441867;
pub const UDMABUF_CREATE: u32 = 2149086530;
pub const SONET_CLRDIAG: u32 = 3221512467;
pub const PHN_SET_REG: u32 = 2147774465;
pub const RNDADDTOENTCNT: u32 = 2147766785;
pub const VBG_IOCTL_CHECK_BALLOON: u32 = 3223344657;
pub const VIDIOC_OMAP3ISP_STAT_REQ: u32 = 3222820550;
pub const PPS_FETCH: u32 = 3221516452;
pub const RTC_AIE_OFF: u32 = 536899586;
pub const VFIO_GROUP_SET_CONTAINER: u32 = 536886120;
pub const FW_CDEV_IOC_RECEIVE_PHY_PACKETS: u32 = 2148016918;
pub const VFIO_IOMMU_SPAPR_TCE_REMOVE: u32 = 536886136;
pub const VFIO_IOMMU_GET_INFO: u32 = 536886128;
pub const DM_DEV_SUSPEND: u32 = 3241737478;
pub const F2FS_IOC_GET_COMPRESS_OPTION: u32 = 1073935637;
pub const FW_CDEV_IOC_STOP_ISO: u32 = 2147754763;
pub const GPIO_V2_GET_LINEINFO_IOCTL: u32 = 3238048773;
pub const ATMMPC_CTRL: u32 = 536895960;
pub const PPPIOCSXASYNCMAP: u32 = 2149610575;
pub const CHIOGSTATUS: u32 = 2148033288;
pub const FW_CDEV_IOC_ALLOCATE_ISO_RESOURCE: u32 = 3222807309;
pub const RIO_MPORT_MAINT_PORT_IDX_GET: u32 = 1074031875;
pub const CAPI_SET_FLAGS: u32 = 1074021156;
pub const VFIO_GROUP_GET_DEVICE_FD: u32 = 536886122;
pub const VHOST_SET_MEM_TABLE: u32 = 2148052739;
pub const MATROXFB_SET_OUTPUT_CONNECTION: u32 = 2147774200;
pub const DFL_FPGA_PORT_GET_REGION_INFO: u32 = 536917570;
pub const VHOST_GET_FEATURES: u32 = 1074310912;
pub const LIRC_GET_REC_RESOLUTION: u32 = 1074030855;
pub const PACKET_CTRL_CMD: u32 = 3222820865;
pub const LIRC_SET_TRANSMITTER_MASK: u32 = 2147772695;
pub const BTRFS_IOC_ADD_DEV: u32 = 2415957002;
pub const JSIOCGCORR: u32 = 1076128290;
pub const VIDIOC_G_FMT: u32 = 3234616836;
pub const RTC_EPOCH_SET: u32 = 2147774478;
pub const CAPI_GET_PROFILE: u32 = 3225436937;
pub const ATM_GETLOOP: u32 = 2148294994;
pub const SCIF_LISTEN: u32 = 2147775234;
pub const NBD_CLEAR_QUE: u32 = 536914693;
pub const F2FS_IOC_MOVE_RANGE: u32 = 3223385353;
pub const LIRC_GET_LENGTH: u32 = 1074030863;
pub const I8K_SET_FAN: u32 = 3221514631;
pub const FDSETMAXERRS: u32 = 2148794956;
pub const VIDIOC_SUBDEV_QUERYCAP: u32 = 1077958144;
pub const SNAPSHOT_SET_SWAP_AREA: u32 = 2148283149;
pub const LIRC_GET_REC_TIMEOUT: u32 = 1074030884;
pub const EVIOCRMFF: u32 = 2147763585;
pub const GPIO_GET_LINEEVENT_IOCTL: u32 = 3224417284;
pub const PPRDATA: u32 = 1073836165;
pub const RIO_MPORT_GET_PROPERTIES: u32 = 1076915460;
pub const TUNSETVNETHDRSZ: u32 = 2147767512;
pub const GPIO_GET_LINEINFO_IOCTL: u32 = 3225990146;
pub const GSMIOC_GETCONF: u32 = 1078740736;
pub const LIRC_GET_SEND_MODE: u32 = 1074030849;
pub const PPPIOCSACTIVE: u32 = 2148037702;
pub const SIOCGSTAMPNS_NEW: u32 = 1074825479;
pub const IPMICTL_RECEIVE_MSG: u32 = 3222825228;
pub const LIRC_SET_SEND_DUTY_CYCLE: u32 = 2147772693;
pub const UI_END_FF_ERASE: u32 = 2148292043;
pub const SWITCHTEC_IOCTL_FLASH_PART_INFO: u32 = 3222296385;
pub const FW_CDEV_IOC_SEND_PHY_PACKET: u32 = 3222807317;
pub const NBD_SET_FLAGS: u32 = 536914698;
pub const SIOCGNETADDR: u32 = 1074039265;
pub const VFIO_DEVICE_GET_REGION_INFO: u32 = 536886124;
pub const REISERFS_IOC_UNPACK: u32 = 2147798273;
pub const FW_CDEV_IOC_REMOVE_DESCRIPTOR: u32 = 2147754759;
pub const RIO_SET_EVENT_MASK: u32 = 2147773709;
pub const SNAPSHOT_ALLOC_SWAP_PAGE: u32 = 1074279188;
pub const VDUSE_VQ_INJECT_IRQ: u32 = 2147778839;
pub const I2OPASSTHRU: u32 = 1074293004;
pub const IOC_OPAL_SET_PW: u32 = 2183164128;
pub const FSI_SCOM_READ: u32 = 3223352065;
pub const VHOST_VDPA_GET_DEVICE_ID: u32 = 1074048880;
pub const VIDIOC_QBUF: u32 = 3225703951;
pub const VIDIOC_S_TUNER: u32 = 2153010718;
pub const TUNGETVNETHDRSZ: u32 = 1074025687;
pub const CAPI_NCCI_GETUNIT: u32 = 1074021159;
pub const DFL_FPGA_PORT_UINT_GET_IRQ_NUM: u32 = 1074050631;
pub const VIDIOC_OMAP3ISP_STAT_EN: u32 = 3221509831;
pub const GPIO_V2_LINE_SET_CONFIG_IOCTL: u32 = 3239097357;
pub const TEE_IOC_VERSION: u32 = 1074570240;
pub const VIDIOC_LOG_STATUS: u32 = 536892998;
pub const IPMICTL_SEND_COMMAND_SETTIME: u32 = 1075603733;
pub const VHOST_SET_LOG_FD: u32 = 2147790599;
pub const SCIF_SEND: u32 = 3222827782;
pub const VIDIOC_SUBDEV_G_FMT: u32 = 3227014660;
pub const NS_ADJBUFLEV: u32 = 536895843;
pub const VIDIOC_DBG_S_REGISTER: u32 = 2151175759;
pub const NILFS_IOCTL_RESIZE: u32 = 2148036235;
pub const PHN_GETREG: u32 = 3221778437;
pub const I2OSWDL: u32 = 3223087365;
pub const VBG_IOCTL_VMMDEV_REQUEST_BIG: u32 = 536892931;
pub const OSIOCGNETADDR: u32 = 1074039265;
pub const JSIOCGBUTTONS: u32 = 1073834514;
pub const VFIO_IOMMU_ENABLE: u32 = 536886131;
pub const DM_DEV_RENAME: u32 = 3241737477;
pub const MEDIA_IOC_SETUP_LINK: u32 = 3224665091;
pub const VIDIOC_ENUMOUTPUT: u32 = 3225966128;
pub const STP_POLICY_ID_SET: u32 = 3222283520;
pub const VHOST_VDPA_SET_CONFIG_CALL: u32 = 2147790711;
pub const VIDIOC_SUBDEV_G_CROP: u32 = 3224917563;
pub const VIDIOC_S_CROP: u32 = 2148816444;
pub const WDIOC_GETTEMP: u32 = 1074026243;
pub const IOC_OPAL_ADD_USR_TO_LR: u32 = 2165862628;
pub const UI_SET_LEDBIT: u32 = 2147767657;
pub const NBD_SET_SOCK: u32 = 536914688;
pub const BTRFS_IOC_SNAP_DESTROY_V2: u32 = 2415957055;
pub const HIDIOCGCOLLECTIONINFO: u32 = 3222292497;
pub const I2OSWUL: u32 = 3223087366;
pub const IOCTL_MEI_NOTIFY_GET: u32 = 1074022403;
pub const FDFMTTRK: u32 = 2148270664;
pub const MMTIMER_GETBITS: u32 = 536898820;
pub const VIDIOC_ENUMSTD: u32 = 3225966105;
pub const VHOST_GET_VRING_BASE: u32 = 3221794578;
pub const VFIO_DEVICE_IOEVENTFD: u32 = 536886132;
pub const ATMARP_SETENTRY: u32 = 536895971;
pub const CCISS_REVALIDVOLS: u32 = 536887818;
pub const MGSL_IOCLOOPTXDONE: u32 = 536898825;
pub const RTC_VL_READ: u32 = 1074032659;
pub const ND_IOCTL_ARS_STATUS: u32 = 3224391171;
pub const RIO_DEV_DEL: u32 = 2149608728;
pub const VBG_IOCTL_ACQUIRE_GUEST_CAPABILITIES: u32 = 3223606797;
pub const VIDIOC_SUBDEV_DV_TIMINGS_CAP: u32 = 3230684772;
pub const SONYPI_IOCSFAN: u32 = 2147579403;
pub const SPIOCSTYPE: u32 = 2147774721;
pub const IPMICTL_REGISTER_FOR_CMD: u32 = 1073899790;
pub const I8K_GET_FAN: u32 = 3221514630;
pub const TUNGETVNETBE: u32 = 1074025695;
pub const AUTOFS_DEV_IOCTL_FAIL: u32 = 3222836087;
pub const UI_END_FF_UPLOAD: u32 = 2153797065;
pub const TOSH_SMM: u32 = 3222828176;
pub const SONYPI_IOCGBAT2REM: u32 = 1073903109;
pub const F2FS_IOC_GET_COMPRESS_BLOCKS: u32 = 1074328849;
pub const PPPIOCSNPMODE: u32 = 2148037707;
pub const USBDEVFS_CONTROL: u32 = 3222295808;
pub const HIDIOCGUSAGE: u32 = 3222816779;
pub const TUNSETTXFILTER: u32 = 2147767505;
pub const TUNGETVNETLE: u32 = 1074025693;
pub const VIDIOC_ENUM_DV_TIMINGS: u32 = 3230946914;
pub const BTRFS_IOC_INO_PATHS: u32 = 3224933411;
pub const MGSL_IOCGXSYNC: u32 = 536898836;
pub const HIDIOCGFIELDINFO: u32 = 3224913930;
pub const VIDIOC_SUBDEV_G_STD: u32 = 1074288151;
pub const I2OVALIDATE: u32 = 1074030856;
pub const VIDIOC_TRY_ENCODER_CMD: u32 = 3223869006;
pub const NILFS_IOCTL_GET_CPINFO: u32 = 1075342978;
pub const VIDIOC_G_FREQUENCY: u32 = 3224131128;
pub const VFAT_IOCTL_READDIR_SHORT: u32 = 1108898306;
pub const ND_IOCTL_GET_CONFIG_DATA: u32 = 3222031877;
pub const F2FS_IOC_RESERVE_COMPRESS_BLOCKS: u32 = 1074328851;
pub const FDGETDRVSTAT: u32 = 1077150226;
pub const SYNC_IOC_MERGE: u32 = 3224387075;
pub const VIDIOC_S_DV_TIMINGS: u32 = 3229898327;
pub const PPPIOCBRIDGECHAN: u32 = 2147775541;
pub const LIRC_SET_SEND_MODE: u32 = 2147772689;
pub const RIO_ENABLE_PORTWRITE_RANGE: u32 = 2148560139;
pub const ATM_GETTYPE: u32 = 2148295044;
pub const PHN_GETREGS: u32 = 3223875591;
pub const FDSETEMSGTRESH: u32 = 536871498;
pub const NILFS_IOCTL_GET_VINFO: u32 = 3222826630;
pub const MGSL_IOCWAITEVENT: u32 = 3221515528;
pub const CAPI_INSTALLED: u32 = 1073890082;
pub const EVIOCGMASK: u32 = 1074808210;
pub const BTRFS_IOC_SUBVOL_GETFLAGS: u32 = 1074304025;
pub const FSL_HV_IOCTL_PARTITION_GET_STATUS: u32 = 3222056706;
pub const MEDIA_IOC_ENUM_ENTITIES: u32 = 3238034433;
pub const GSMIOC_GETFIRST: u32 = 1074022148;
pub const FW_CDEV_IOC_FLUSH_ISO: u32 = 2147754776;
pub const VIDIOC_DBG_G_CHIP_INFO: u32 = 3234354790;
pub const F2FS_IOC_RELEASE_VOLATILE_WRITE: u32 = 536933636;
pub const CAPI_GET_SERIAL: u32 = 3221504776;
pub const FDSETDRVPRM: u32 = 2153251472;
pub const IOC_OPAL_SAVE: u32 = 2165862620;
pub const VIDIOC_G_DV_TIMINGS: u32 = 3229898328;
pub const TUNSETIFINDEX: u32 = 2147767514;
pub const CCISS_SETINTINFO: u32 = 2148024835;
pub const CM_IOSDBGLVL: u32 = 2147771386;
pub const RTC_VL_CLR: u32 = 536899604;
pub const VIDIOC_REQBUFS: u32 = 3222558216;
pub const USBDEVFS_REAPURBNDELAY32: u32 = 2147767565;
pub const TEE_IOC_SHM_REGISTER: u32 = 3222840329;
pub const USBDEVFS_SETCONFIGURATION: u32 = 1074025733;
pub const CCISS_GETNODENAME: u32 = 1074807300;
pub const VIDIOC_SUBDEV_S_FRAME_INTERVAL: u32 = 3224393238;
pub const VIDIOC_ENUM_FRAMESIZES: u32 = 3224131146;
pub const VFIO_DEVICE_PCI_HOT_RESET: u32 = 536886129;
pub const FW_CDEV_IOC_SEND_BROADCAST_REQUEST: u32 = 2150114066;
pub const LPSETTIMEOUT_NEW: u32 = 2148533775;
pub const RIO_CM_MPORT_GET_LIST: u32 = 3221512971;
pub const FW_CDEV_IOC_QUEUE_ISO: u32 = 3222807305;
pub const FDRAWCMD: u32 = 536871512;
pub const SCIF_UNREG: u32 = 3222303497;
pub const PPPIOCGIDLE64: u32 = 1074820159;
pub const USBDEVFS_RELEASEINTERFACE: u32 = 1074025744;
pub const VIDIOC_CROPCAP: u32 = 3224131130;
pub const DFL_FPGA_PORT_GET_INFO: u32 = 536917569;
pub const PHN_SET_REGS: u32 = 2147774467;
pub const ATMLEC_DATA: u32 = 536895953;
pub const PPPOEIOCDFWD: u32 = 536916225;
pub const VIDIOC_S_SELECTION: u32 = 3225441887;
pub const SNAPSHOT_FREE_SWAP_PAGES: u32 = 536883977;
pub const BTRFS_IOC_LOGICAL_INO: u32 = 3224933412;
pub const VIDIOC_S_CTRL: u32 = 3221771804;
pub const ZATM_SETPOOL: u32 = 2148295011;
pub const MTIOCPOS: u32 = 1074031875;
pub const PMU_IOC_SLEEP: u32 = 536887808;
pub const AUTOFS_DEV_IOCTL_PROTOSUBVER: u32 = 3222836083;
pub const VBG_IOCTL_CHANGE_FILTER_MASK: u32 = 3223344652;
pub const NILFS_IOCTL_GET_SUSTAT: u32 = 1076915845;
pub const VIDIOC_QUERYCAP: u32 = 1080579584;
pub const HPET_INFO: u32 = 1074554883;
pub const VIDIOC_AM437X_CCDC_CFG: u32 = 2147768001;
pub const DM_LIST_DEVICES: u32 = 3241737474;
pub const TUNSETOWNER: u32 = 2147767500;
pub const VBG_IOCTL_CHANGE_GUEST_CAPABILITIES: u32 = 3223344654;
pub const RNDADDENTROPY: u32 = 2148028931;
pub const USBDEVFS_RESET: u32 = 536892692;
pub const BTRFS_IOC_SUBVOL_CREATE: u32 = 2415957006;
pub const USBDEVFS_FORBID_SUSPEND: u32 = 536892705;
pub const FDGETDRVTYP: u32 = 1074790927;
pub const PPWCONTROL: u32 = 2147577988;
pub const VIDIOC_ENUM_FRAMEINTERVALS: u32 = 3224655435;
pub const KCOV_DISABLE: u32 = 536896357;
pub const IOC_OPAL_ACTIVATE_LSP: u32 = 2165862623;
pub const VHOST_VDPA_GET_IOVA_RANGE: u32 = 1074835320;
pub const PPPIOCSPASS: u32 = 2148037703;
pub const RIO_CM_CHAN_CONNECT: u32 = 2148033288;
pub const I2OSWDEL: u32 = 3223087367;
pub const FS_IOC_SET_ENCRYPTION_POLICY: u32 = 1074554387;
pub const IOC_OPAL_MBR_DONE: u32 = 2165338345;
pub const PPPIOCSMAXCID: u32 = 2147775569;
pub const PPSETPHASE: u32 = 2147774612;
pub const VHOST_VDPA_SET_VRING_ENABLE: u32 = 2148052853;
pub const USBDEVFS_GET_SPEED: u32 = 536892703;
pub const SONET_GETFRAMING: u32 = 1074028822;
pub const VIDIOC_QUERYBUF: u32 = 3225703945;
pub const VIDIOC_S_EDID: u32 = 3223606825;
pub const BTRFS_IOC_QGROUP_ASSIGN: u32 = 2149094441;
pub const PPS_GETCAP: u32 = 1074032803;
pub const SNAPSHOT_PLATFORM_SUPPORT: u32 = 536883983;
pub const LIRC_SET_REC_TIMEOUT_REPORTS: u32 = 2147772697;
pub const SCIF_GET_NODEIDS: u32 = 3222827790;
pub const NBD_DISCONNECT: u32 = 536914696;
pub const VIDIOC_SUBDEV_G_FRAME_INTERVAL: u32 = 3224393237;
pub const VFIO_IOMMU_DISABLE: u32 = 536886132;
pub const SNAPSHOT_CREATE_IMAGE: u32 = 2147758865;
pub const SNAPSHOT_POWER_OFF: u32 = 536883984;
pub const APM_IOC_STANDBY: u32 = 536887553;
pub const PPPIOCGUNIT: u32 = 1074033750;
pub const AUTOFS_IOC_EXPIRE_MULTI: u32 = 2147783526;
pub const SCIF_BIND: u32 = 3221779201;
pub const IOC_WATCH_QUEUE_SET_SIZE: u32 = 536893280;
pub const NILFS_IOCTL_CHANGE_CPMODE: u32 = 2148560512;
pub const IOC_OPAL_LOCK_UNLOCK: u32 = 2165862621;
pub const F2FS_IOC_SET_PIN_FILE: u32 = 2147808525;
pub const PPPIOCGRASYNCMAP: u32 = 1074033749;
pub const MMTIMER_MMAPAVAIL: u32 = 536898822;
pub const I2OPASSTHRU32: u32 = 1074293004;
pub const DFL_FPGA_FME_PORT_RELEASE: u32 = 2147792513;
pub const VIDIOC_SUBDEV_QUERY_DV_TIMINGS: u32 = 1082414691;
pub const UI_SET_SNDBIT: u32 = 2147767658;
pub const VIDIOC_G_AUDOUT: u32 = 1077171761;
pub const RTC_PLL_SET: u32 = 2149347346;
pub const VIDIOC_ENUMAUDIO: u32 = 3224655425;
pub const AUTOFS_DEV_IOCTL_TIMEOUT: u32 = 3222836090;
pub const VBG_IOCTL_DRIVER_VERSION_INFO: u32 = 3224131072;
pub const VHOST_SCSI_GET_EVENTS_MISSED: u32 = 2147790660;
pub const VHOST_SET_VRING_ADDR: u32 = 2150149905;
pub const VDUSE_CREATE_DEV: u32 = 2169536770;
pub const FDFLUSH: u32 = 536871499;
pub const VBG_IOCTL_WAIT_FOR_EVENTS: u32 = 3223344650;
pub const DFL_FPGA_FME_ERR_SET_IRQ: u32 = 2148054660;
pub const F2FS_IOC_GET_PIN_FILE: u32 = 1074066702;
pub const SCIF_CONNECT: u32 = 3221779203;
pub const BLKREPORTZONE: u32 = 3222278786;
pub const AUTOFS_IOC_ASKUMOUNT: u32 = 1074041712;
pub const ATM_ADDPARTY: u32 = 2148033012;
pub const FDSETPRM: u32 = 2149319234;
pub const ATM_GETSTATZ: u32 = 2148294993;
pub const ISST_IF_MSR_COMMAND: u32 = 3221552644;
pub const BTRFS_IOC_GET_SUBVOL_INFO: u32 = 1106809916;
pub const VIDIOC_UNSUBSCRIBE_EVENT: u32 = 2149602907;
pub const SEV_ISSUE_CMD: u32 = 3222295296;
pub const GPIOHANDLE_SET_LINE_VALUES_IOCTL: u32 = 3225465865;
pub const PCITEST_COPY: u32 = 2147766278;
pub const IPMICTL_GET_MY_ADDRESS_CMD: u32 = 1074030866;
pub const CHIOGPICKER: u32 = 1074029316;
pub const CAPI_NCCI_OPENCOUNT: u32 = 1074021158;
pub const CXL_MEM_SEND_COMMAND: u32 = 3224423938;
pub const PERF_EVENT_IOC_SET_FILTER: u32 = 2147755014;
pub const IOC_OPAL_REVERT_TPR: u32 = 2164814050;
pub const CHIOGVPARAMS: u32 = 1081107219;
pub const PTP_PEROUT_REQUEST: u32 = 2151169283;
pub const FSI_SCOM_CHECK: u32 = 1074033408;
pub const RTC_IRQP_READ: u32 = 1074032651;
pub const RIO_MPORT_MAINT_READ_LOCAL: u32 = 1075342597;
pub const HIDIOCGRDESCSIZE: u32 = 1074022401;
pub const UI_GET_VERSION: u32 = 1074025773;
pub const NILFS_IOCTL_GET_CPSTAT: u32 = 1075342979;
pub const CCISS_GETBUSTYPES: u32 = 1074020871;
pub const VFIO_IOMMU_SPAPR_TCE_CREATE: u32 = 536886135;
pub const VIDIOC_EXPBUF: u32 = 3225441808;
pub const UI_SET_RELBIT: u32 = 2147767654;
pub const VFIO_SET_IOMMU: u32 = 536886118;
pub const VIDIOC_S_MODULATOR: u32 = 2151962167;
pub const TUNGETFILTER: u32 = 1074287835;
pub const MEYEIOC_SYNC: u32 = 3221518019;
pub const CCISS_SETNODENAME: u32 = 2148549125;
pub const FBIO_GETCONTROL2: u32 = 1074022025;
pub const TUNSETDEBUG: u32 = 2147767497;
pub const DM_DEV_REMOVE: u32 = 3241737476;
pub const HIDIOCSUSAGES: u32 = 2417772564;
pub const FS_IOC_ADD_ENCRYPTION_KEY: u32 = 3226494487;
pub const FBIOGET_VBLANK: u32 = 1075856914;
pub const ATM_GETSTAT: u32 = 2148294992;
pub const VIDIOC_G_JPEGCOMP: u32 = 1082938941;
pub const TUNATTACHFILTER: u32 = 2148029653;
pub const UI_SET_ABSBIT: u32 = 2147767655;
pub const DFL_FPGA_PORT_ERR_GET_IRQ_NUM: u32 = 1074050629;
pub const USBDEVFS_REAPURB32: u32 = 2147767564;
pub const BTRFS_IOC_TRANS_END: u32 = 536908807;
pub const CAPI_REGISTER: u32 = 2148287233;
pub const F2FS_IOC_COMPRESS_FILE: u32 = 536933656;
pub const USBDEVFS_DISCARDURB: u32 = 536892683;
pub const HE_GET_REG: u32 = 2148295008;
pub const ATM_SETLOOP: u32 = 2148294995;
pub const ATMSIGD_CTRL: u32 = 536895984;
pub const CIOC_KERNEL_VERSION: u32 = 3221512970;
pub const BTRFS_IOC_CLONE_RANGE: u32 = 2149618701;
pub const SNAPSHOT_UNFREEZE: u32 = 536883970;
pub const F2FS_IOC_START_VOLATILE_WRITE: u32 = 536933635;
pub const PMU_IOC_HAS_ADB: u32 = 1074020868;
pub const I2OGETIOPS: u32 = 1075865856;
pub const VIDIOC_S_FBUF: u32 = 2150389259;
pub const PPRCONTROL: u32 = 1073836163;
pub const CHIOSPICKER: u32 = 2147771141;
pub const VFIO_IOMMU_SPAPR_REGISTER_MEMORY: u32 = 536886133;
pub const TUNGETSNDBUF: u32 = 1074025683;
pub const GSMIOC_SETCONF: u32 = 2152482561;
pub const IOC_PR_PREEMPT: u32 = 2149085387;
pub const KCOV_INIT_TRACE: u32 = 1074029313;
pub const SONYPI_IOCGBAT1CAP: u32 = 1073903106;
pub const SWITCHTEC_IOCTL_FLASH_INFO: u32 = 1074812736;
pub const MTIOCTOP: u32 = 2148035841;
pub const VHOST_VDPA_SET_STATUS: u32 = 2147594098;
pub const VHOST_SCSI_SET_EVENTS_MISSED: u32 = 2147790659;
pub const VFIO_IOMMU_DIRTY_PAGES: u32 = 536886133;
pub const BTRFS_IOC_SCRUB_PROGRESS: u32 = 3288372253;
pub const PPPIOCGMRU: u32 = 1074033747;
pub const BTRFS_IOC_DEV_REPLACE: u32 = 3391657013;
pub const PPPIOCGFLAGS: u32 = 1074033754;
pub const NILFS_IOCTL_SET_SUINFO: u32 = 2149084813;
pub const FW_CDEV_IOC_GET_CYCLE_TIMER2: u32 = 3222807316;
pub const ATM_DELLECSADDR: u32 = 2148295055;
pub const FW_CDEV_IOC_GET_SPEED: u32 = 536879889;
pub const PPPIOCGIDLE32: u32 = 1074295871;
pub const VFIO_DEVICE_RESET: u32 = 536886127;
pub const GPIO_GET_LINEINFO_UNWATCH_IOCTL: u32 = 3221533708;
pub const WDIOC_GETSTATUS: u32 = 1074026241;
pub const BTRFS_IOC_SET_FEATURES: u32 = 2150667321;
pub const IOCTL_MEI_CONNECT_CLIENT: u32 = 3222292481;
pub const VIDIOC_OMAP3ISP_AEWB_CFG: u32 = 3223344835;
pub const PCITEST_READ: u32 = 2147766277;
pub const VFIO_GROUP_GET_STATUS: u32 = 536886119;
pub const MATROXFB_GET_ALL_OUTPUTS: u32 = 1074032379;
pub const USBDEVFS_CLEAR_HALT: u32 = 1074025749;
pub const VIDIOC_DECODER_CMD: u32 = 3225966176;
pub const VIDIOC_G_AUDIO: u32 = 1077171745;
pub const CCISS_RESCANDISK: u32 = 536887824;
pub const RIO_DISABLE_PORTWRITE_RANGE: u32 = 2148560140;
pub const IOC_OPAL_SECURE_ERASE_LR: u32 = 2165338343;
pub const USBDEVFS_REAPURB: u32 = 2147767564;
pub const DFL_FPGA_CHECK_EXTENSION: u32 = 536917505;
pub const AUTOFS_IOC_PROTOVER: u32 = 1074041699;
pub const FSL_HV_IOCTL_MEMCPY: u32 = 3223891717;
pub const BTRFS_IOC_GET_FEATURES: u32 = 1075352633;
pub const PCITEST_MSIX: u32 = 2147766279;
pub const BTRFS_IOC_DEFRAG_RANGE: u32 = 2150667280;
pub const UI_BEGIN_FF_ERASE: u32 = 3222033866;
pub const DM_GET_TARGET_VERSION: u32 = 3241737489;
pub const PPPIOCGIDLE: u32 = 1074295871;
pub const NVRAM_SETCKS: u32 = 536899649;
pub const WDIOC_GETSUPPORT: u32 = 1076385536;
pub const GSMIOC_ENABLE_NET: u32 = 2150909698;
pub const GPIO_GET_CHIPINFO_IOCTL: u32 = 1078244353;
pub const NE_ADD_VCPU: u32 = 3221532193;
pub const EVIOCSKEYCODE_V2: u32 = 2150122756;
pub const PTP_SYS_OFFSET_EXTENDED2: u32 = 3300932882;
pub const SCIF_FENCE_WAIT: u32 = 3221517072;
pub const RIO_TRANSFER: u32 = 3222826261;
pub const FSL_HV_IOCTL_DOORBELL: u32 = 3221794566;
pub const RIO_MPORT_MAINT_WRITE_LOCAL: u32 = 2149084422;
pub const I2OEVTREG: u32 = 2148296970;
pub const I2OPARMGET: u32 = 3222825220;
pub const EVIOCGID: u32 = 1074283778;
pub const BTRFS_IOC_QGROUP_CREATE: u32 = 2148570154;
pub const AUTOFS_DEV_IOCTL_SETPIPEFD: u32 = 3222836088;
pub const VIDIOC_S_PARM: u32 = 3234616854;
pub const TUNSETSTEERINGEBPF: u32 = 1074025696;
pub const ATM_GETNAMES: u32 = 2148032899;
pub const VIDIOC_QUERYMENU: u32 = 3224131109;
pub const DFL_FPGA_PORT_DMA_UNMAP: u32 = 536917572;
pub const I2OLCTGET: u32 = 3222038786;
pub const FS_IOC_GET_ENCRYPTION_PWSALT: u32 = 2148558356;
pub const NS_SETBUFLEV: u32 = 2148295010;
pub const BLKCLOSEZONE: u32 = 2148536967;
pub const SONET_GETFRSENSE: u32 = 1074159895;
pub const UI_SET_EVBIT: u32 = 2147767652;
pub const DM_LIST_VERSIONS: u32 = 3241737485;
pub const HIDIOCGSTRING: u32 = 1090799620;
pub const PPPIOCATTCHAN: u32 = 2147775544;
pub const VDUSE_DEV_SET_CONFIG: u32 = 2148040978;
pub const TUNGETFEATURES: u32 = 1074025679;
pub const VFIO_GROUP_UNSET_CONTAINER: u32 = 536886121;
pub const IPMICTL_SET_MY_ADDRESS_CMD: u32 = 1074030865;
pub const CCISS_REGNEWDISK: u32 = 2147762701;
pub const VIDIOC_QUERY_DV_TIMINGS: u32 = 1082414691;
pub const PHN_SETREGS: u32 = 2150133768;
pub const FAT_IOCTL_GET_ATTRIBUTES: u32 = 1074033168;
pub const FSL_MC_SEND_MC_COMMAND: u32 = 3225440992;
pub const TUNGETIFF: u32 = 1074025682;
pub const PTP_CLOCK_GETCAPS2: u32 = 1079000330;
pub const BTRFS_IOC_RESIZE: u32 = 2415956995;
pub const VHOST_SET_VRING_ENDIAN: u32 = 2148052755;
pub const PPS_KC_BIND: u32 = 2147774629;
pub const F2FS_IOC_WRITE_CHECKPOINT: u32 = 536933639;
pub const UI_SET_FFBIT: u32 = 2147767659;
pub const IPMICTL_GET_MY_LUN_CMD: u32 = 1074030868;
pub const CEC_ADAP_G_PHYS_ADDR: u32 = 1073897729;
pub const CEC_G_MODE: u32 = 1074028808;
pub const USBDEVFS_RESETEP: u32 = 1074025731;
pub const MEDIA_REQUEST_IOC_QUEUE: u32 = 536902784;
pub const USBDEVFS_ALLOC_STREAMS: u32 = 1074287900;
pub const MGSL_IOCSXCTRL: u32 = 536898837;
pub const MEDIA_IOC_G_TOPOLOGY: u32 = 3225975812;
pub const PPPIOCUNBRIDGECHAN: u32 = 536900660;
pub const F2FS_IOC_COMMIT_ATOMIC_WRITE: u32 = 536933634;
pub const ISST_IF_GET_PLATFORM_INFO: u32 = 1074068992;
pub const SCIF_FENCE_MARK: u32 = 3222303503;
pub const USBDEVFS_RELEASE_PORT: u32 = 1074025753;
pub const VFIO_CHECK_EXTENSION: u32 = 536886117;
pub const BTRFS_IOC_QGROUP_LIMIT: u32 = 1076925483;
pub const FAT_IOCTL_GET_VOLUME_ID: u32 = 1074033171;
pub const UI_SET_PHYS: u32 = 2147767660;
pub const FDWERRORGET: u32 = 1075315223;
pub const VIDIOC_SUBDEV_G_EDID: u32 = 3223606824;
pub const MGSL_IOCGSTATS: u32 = 536898823;
pub const RPROC_SET_SHUTDOWN_ON_RELEASE: u32 = 2147792641;
pub const SIOCGSTAMP_NEW: u32 = 1074825478;
pub const RTC_WKALM_RD: u32 = 1076391952;
pub const PHN_GET_REG: u32 = 3221516288;
pub const DELL_WMI_SMBIOS_CMD: u32 = 3224655616;
pub const PHN_NOT_OH: u32 = 536899588;
pub const PPGETMODES: u32 = 1074032791;
pub const CHIOGPARAMS: u32 = 1075077894;
pub const VFIO_DEVICE_GET_GFX_DMABUF: u32 = 536886131;
pub const VHOST_SET_VRING_BUSYLOOP_TIMEOUT: u32 = 2148052771;
pub const VIDIOC_SUBDEV_G_SELECTION: u32 = 3225441853;
pub const BTRFS_IOC_RM_DEV_V2: u32 = 2415957050;
pub const MGSL_IOCWAITGPIO: u32 = 3222301970;
pub const PMU_IOC_CAN_SLEEP: u32 = 1074020869;
pub const KCOV_ENABLE: u32 = 536896356;
pub const BTRFS_IOC_CLONE: u32 = 2147783689;
pub const F2FS_IOC_DEFRAGMENT: u32 = 3222336776;
pub const FW_CDEV_IOC_DEALLOCATE_ISO_RESOURCE: u32 = 2147754766;
pub const AGPIOC_ALLOCATE: u32 = 3221504262;
pub const NE_SET_USER_MEMORY_REGION: u32 = 2149101091;
pub const MGSL_IOCTXABORT: u32 = 536898822;
pub const MGSL_IOCSGPIO: u32 = 2148560144;
pub const LIRC_SET_REC_CARRIER: u32 = 2147772692;
pub const F2FS_IOC_FLUSH_DEVICE: u32 = 2148070666;
pub const SNAPSHOT_ATOMIC_RESTORE: u32 = 536883972;
pub const RTC_UIE_OFF: u32 = 536899588;
pub const BT_BMC_IOCTL_SMS_ATN: u32 = 536916224;
pub const NVME_IOCTL_ID: u32 = 536890944;
pub const NE_START_ENCLAVE: u32 = 3222318628;
pub const VIDIOC_STREAMON: u32 = 2147767826;
pub const FDPOLLDRVSTAT: u32 = 1077150227;
pub const AUTOFS_DEV_IOCTL_READY: u32 = 3222836086;
pub const VIDIOC_ENUMAUDOUT: u32 = 3224655426;
pub const VIDIOC_SUBDEV_S_STD: u32 = 2148029976;
pub const WDIOC_GETTIMELEFT: u32 = 1074026250;
pub const ATM_GETLINKRATE: u32 = 2148295041;
pub const RTC_WKALM_SET: u32 = 2150133775;
pub const VHOST_GET_BACKEND_FEATURES: u32 = 1074310950;
pub const ATMARP_ENCAP: u32 = 536895973;
pub const CAPI_GET_FLAGS: u32 = 1074021155;
pub const IPMICTL_SET_MY_CHANNEL_ADDRESS_CMD: u32 = 1074030872;
pub const DFL_FPGA_FME_PORT_ASSIGN: u32 = 2147792514;
pub const NS_GET_OWNER_UID: u32 = 536917764;
pub const VIDIOC_OVERLAY: u32 = 2147767822;
pub const BTRFS_IOC_WAIT_SYNC: u32 = 2148045846;
pub const GPIOHANDLE_SET_CONFIG_IOCTL: u32 = 3226776586;
pub const VHOST_GET_VRING_ENDIAN: u32 = 2148052756;
pub const ATM_GETADDR: u32 = 2148295046;
pub const PHN_GET_REGS: u32 = 3221516290;
pub const AUTOFS_DEV_IOCTL_REQUESTER: u32 = 3222836091;
pub const AUTOFS_DEV_IOCTL_EXPIRE: u32 = 3222836092;
pub const SNAPSHOT_S2RAM: u32 = 536883979;
pub const JSIOCSAXMAP: u32 = 2151705137;
pub const F2FS_IOC_SET_COMPRESS_OPTION: u32 = 2147677462;
pub const VBG_IOCTL_HGCM_DISCONNECT: u32 = 3223082501;
pub const SCIF_FENCE_SIGNAL: u32 = 3223876369;
pub const VFIO_DEVICE_GET_PCI_HOT_RESET_INFO: u32 = 536886128;
pub const VIDIOC_SUBDEV_ENUM_MBUS_CODE: u32 = 3224393218;
pub const MMTIMER_GETOFFSET: u32 = 536898816;
pub const RIO_CM_CHAN_LISTEN: u32 = 2147640070;
pub const ATM_SETSC: u32 = 2147770865;
pub const F2FS_IOC_SHUTDOWN: u32 = 1074026621;
pub const NVME_IOCTL_RESCAN: u32 = 536890950;
pub const BLKOPENZONE: u32 = 2148536966;
pub const DM_VERSION: u32 = 3241737472;
pub const CEC_TRANSMIT: u32 = 3224920325;
pub const FS_IOC_GET_ENCRYPTION_POLICY_EX: u32 = 3221841430;
pub const SIOCMKCLIP: u32 = 536895968;
pub const IPMI_BMC_IOCTL_CLEAR_SMS_ATN: u32 = 536916225;
pub const HIDIOCGVERSION: u32 = 1074022401;
pub const VIDIOC_S_INPUT: u32 = 3221509671;
pub const VIDIOC_G_CROP: u32 = 3222558267;
pub const LIRC_SET_WIDEBAND_RECEIVER: u32 = 2147772707;
pub const EVIOCGEFFECTS: u32 = 1074021764;
pub const UVCIOC_CTRL_QUERY: u32 = 3222041889;
pub const IOC_OPAL_GENERIC_TABLE_RW: u32 = 2167959787;
pub const FS_IOC_READ_VERITY_METADATA: u32 = 3223873159;
pub const ND_IOCTL_SET_CONFIG_DATA: u32 = 3221769734;
pub const USBDEVFS_GETDRIVER: u32 = 2164544776;
pub const IDT77105_GETSTAT: u32 = 2148294962;
pub const HIDIOCINITREPORT: u32 = 536889349;
pub const VFIO_DEVICE_GET_INFO: u32 = 536886123;
pub const RIO_CM_CHAN_RECEIVE: u32 = 3222299402;
pub const RNDGETENTCNT: u32 = 1074024960;
pub const PPPIOCNEWUNIT: u32 = 3221517374;
pub const BTRFS_IOC_INO_LOOKUP: u32 = 3489698834;
pub const FDRESET: u32 = 536871508;
pub const IOC_PR_REGISTER: u32 = 2149085384;
pub const HIDIOCSREPORT: u32 = 2148288520;
pub const TEE_IOC_OPEN_SESSION: u32 = 1074832386;
pub const TEE_IOC_SUPPL_RECV: u32 = 1074832390;
pub const BTRFS_IOC_BALANCE_CTL: u32 = 2147783713;
pub const GPIO_GET_LINEINFO_WATCH_IOCTL: u32 = 3225990155;
pub const HIDIOCGRAWINFO: u32 = 1074284547;
pub const PPPIOCSCOMPRESS: u32 = 2148299853;
pub const USBDEVFS_CONNECTINFO: u32 = 2148029713;
pub const BLKRESETZONE: u32 = 2148536963;
pub const CHIOINITELEM: u32 = 536896273;
pub const NILFS_IOCTL_SET_ALLOC_RANGE: u32 = 2148560524;
pub const AUTOFS_DEV_IOCTL_CATATONIC: u32 = 3222836089;
pub const RIO_MPORT_MAINT_HDID_SET: u32 = 2147642625;
pub const PPGETPHASE: u32 = 1074032793;
pub const USBDEVFS_DISCONNECT_CLAIM: u32 = 1091065115;
pub const FDMSGON: u32 = 536871493;
pub const VIDIOC_G_SLICED_VBI_CAP: u32 = 3228849733;
pub const BTRFS_IOC_BALANCE_V2: u32 = 3288372256;
pub const MEDIA_REQUEST_IOC_REINIT: u32 = 536902785;
pub const IOC_OPAL_ERASE_LR: u32 = 2165338342;
pub const FDFMTBEG: u32 = 536871495;
pub const RNDRESEEDCRNG: u32 = 536891911;
pub const ISST_IF_GET_PHY_ID: u32 = 3221552641;
pub const TUNSETNOCSUM: u32 = 2147767496;
pub const SONET_GETSTAT: u32 = 1076125968;
pub const TFD_IOC_SET_TICKS: u32 = 2148029440;
pub const PPDATADIR: u32 = 2147774608;
pub const IOC_OPAL_ENABLE_DISABLE_MBR: u32 = 2165338341;
pub const GPIO_V2_GET_LINE_IOCTL: u32 = 3260068871;
pub const RIO_CM_CHAN_SEND: u32 = 2148557577;
pub const PPWCTLONIRQ: u32 = 2147578002;
pub const SONYPI_IOCGBRT: u32 = 1073837568;
pub const IOC_PR_RELEASE: u32 = 2148561098;
pub const PPCLRIRQ: u32 = 1074032787;
pub const IPMICTL_SET_MY_CHANNEL_LUN_CMD: u32 = 1074030874;
pub const MGSL_IOCSXSYNC: u32 = 536898835;
pub const HPET_IE_OFF: u32 = 536897538;
pub const IOC_OPAL_ACTIVATE_USR: u32 = 2165338337;
pub const SONET_SETFRAMING: u32 = 2147770645;
pub const PERF_EVENT_IOC_PAUSE_OUTPUT: u32 = 2147755017;
pub const BTRFS_IOC_LOGICAL_INO_V2: u32 = 3224933435;
pub const VBG_IOCTL_HGCM_CONNECT: u32 = 3231471108;
pub const BLKFINISHZONE: u32 = 2148536968;
pub const EVIOCREVOKE: u32 = 2147763601;
pub const VFIO_DEVICE_FEATURE: u32 = 536886133;
pub const CCISS_GETPCIINFO: u32 = 1074283009;
pub const ISST_IF_MBOX_COMMAND: u32 = 3221552643;
pub const SCIF_ACCEPTREQ: u32 = 3222303492;
pub const PERF_EVENT_IOC_QUERY_BPF: u32 = 3221496842;
pub const VIDIOC_STREAMOFF: u32 = 2147767827;
pub const VDUSE_DESTROY_DEV: u32 = 2164293891;
pub const FDGETFDCSTAT: u32 = 1075839509;
pub const CM_IOCGATR: u32 = 3221512961;
pub const VIDIOC_S_PRIORITY: u32 = 2147767876;
pub const SNAPSHOT_FREEZE: u32 = 536883969;
pub const VIDIOC_ENUMINPUT: u32 = 3226490394;
pub const ZATM_GETPOOLZ: u32 = 2148295010;
pub const RIO_DISABLE_DOORBELL_RANGE: u32 = 2148035850;
pub const GPIO_V2_GET_LINEINFO_WATCH_IOCTL: u32 = 3238048774;
pub const VIDIOC_G_STD: u32 = 1074288151;
pub const USBDEVFS_ALLOW_SUSPEND: u32 = 536892706;
pub const SONET_GETSTATZ: u32 = 1076125969;
pub const SCIF_ACCEPTREG: u32 = 3221779205;
pub const VIDIOC_ENCODER_CMD: u32 = 3223869005;
pub const PPPIOCSRASYNCMAP: u32 = 2147775572;
pub const IOCTL_MEI_NOTIFY_SET: u32 = 2147764226;
pub const BTRFS_IOC_QUOTA_RESCAN_STATUS: u32 = 1077974061;
pub const F2FS_IOC_GARBAGE_COLLECT: u32 = 2147808518;
pub const ATMLEC_CTRL: u32 = 536895952;
pub const MATROXFB_GET_AVAILABLE_OUTPUTS: u32 = 1074032377;
pub const DM_DEV_CREATE: u32 = 3241737475;
pub const VHOST_VDPA_GET_VRING_NUM: u32 = 1073917814;
pub const VIDIOC_G_CTRL: u32 = 3221771803;
pub const NBD_CLEAR_SOCK: u32 = 536914692;
pub const VFIO_DEVICE_QUERY_GFX_PLANE: u32 = 536886130;
pub const WDIOC_KEEPALIVE: u32 = 1074026245;
pub const NVME_IOCTL_SUBSYS_RESET: u32 = 536890949;
pub const PTP_EXTTS_REQUEST2: u32 = 2148547851;
pub const PCITEST_BAR: u32 = 536891393;
pub const MGSL_IOCGGPIO: u32 = 1074818321;
pub const EVIOCSREP: u32 = 2148025603;
pub const VFIO_DEVICE_GET_IRQ_INFO: u32 = 536886125;
pub const HPET_DPI: u32 = 536897541;
pub const VDUSE_VQ_SETUP_KICKFD: u32 = 2148040982;
pub const ND_IOCTL_CALL: u32 = 3225439754;
pub const HIDIOCGDEVINFO: u32 = 1075595267;
pub const DM_TABLE_DEPS: u32 = 3241737483;
pub const BTRFS_IOC_DEV_INFO: u32 = 3489698846;
pub const VDUSE_IOTLB_GET_FD: u32 = 3223355664;
pub const FW_CDEV_IOC_GET_INFO: u32 = 3223855872;
pub const VIDIOC_G_PRIORITY: u32 = 1074026051;
pub const ATM_NEWBACKENDIF: u32 = 2147639795;
pub const VIDIOC_S_EXT_CTRLS: u32 = 3222820424;
pub const VIDIOC_SUBDEV_ENUM_DV_TIMINGS: u32 = 3230946914;
pub const VIDIOC_OMAP3ISP_CCDC_CFG: u32 = 3223344833;
pub const VIDIOC_S_HW_FREQ_SEEK: u32 = 2150651474;
pub const DM_TABLE_LOAD: u32 = 3241737481;
pub const F2FS_IOC_START_ATOMIC_WRITE: u32 = 536933633;
pub const VIDIOC_G_OUTPUT: u32 = 1074026030;
pub const ATM_DROPPARTY: u32 = 2147770869;
pub const CHIOGELEM: u32 = 2154586896;
pub const BTRFS_IOC_GET_SUPPORTED_FEATURES: u32 = 1078498361;
pub const EVIOCSKEYCODE: u32 = 2148025604;
pub const NE_GET_IMAGE_LOAD_INFO: u32 = 3222318626;
pub const TUNSETLINK: u32 = 2147767501;
pub const FW_CDEV_IOC_ADD_DESCRIPTOR: u32 = 3222807302;
pub const BTRFS_IOC_SCRUB_CANCEL: u32 = 536908828;
pub const PPS_SETPARAMS: u32 = 2147774626;
pub const IOC_OPAL_LR_SETUP: u32 = 2166911203;
pub const FW_CDEV_IOC_DEALLOCATE: u32 = 2147754755;
pub const WDIOC_SETTIMEOUT: u32 = 3221509894;
pub const IOC_WATCH_QUEUE_SET_FILTER: u32 = 536893281;
pub const CAPI_GET_MANUFACTURER: u32 = 3221504774;
pub const VFIO_IOMMU_SPAPR_UNREGISTER_MEMORY: u32 = 536886134;
pub const ASPEED_P2A_CTRL_IOCTL_SET_WINDOW: u32 = 2148578048;
pub const VIDIOC_G_EDID: u32 = 3223606824;
pub const F2FS_IOC_GARBAGE_COLLECT_RANGE: u32 = 2149119243;
pub const RIO_MAP_INBOUND: u32 = 3223874833;
pub const IOC_OPAL_TAKE_OWNERSHIP: u32 = 2164814046;
pub const USBDEVFS_CLAIM_PORT: u32 = 1074025752;
pub const VIDIOC_S_AUDIO: u32 = 2150913570;
pub const FS_IOC_GET_ENCRYPTION_NONCE: u32 = 1074816539;
pub const FW_CDEV_IOC_SEND_STREAM_PACKET: u32 = 2150114067;
pub const BTRFS_IOC_SNAP_DESTROY: u32 = 2415957007;
pub const SNAPSHOT_FREE: u32 = 536883973;
pub const I8K_GET_SPEED: u32 = 3221514629;
pub const HIDIOCGREPORT: u32 = 2148288519;
pub const HPET_EPI: u32 = 536897540;
pub const JSIOCSCORR: u32 = 2149870113;
pub const IOC_PR_PREEMPT_ABORT: u32 = 2149085388;
pub const RIO_MAP_OUTBOUND: u32 = 3223874831;
pub const ATM_SETESI: u32 = 2148295052;
pub const FW_CDEV_IOC_START_ISO: u32 = 2148541194;
pub const ATM_DELADDR: u32 = 2148295049;
pub const PPFCONTROL: u32 = 2147643534;
pub const SONYPI_IOCGFAN: u32 = 1073837578;
pub const RTC_IRQP_SET: u32 = 2147774476;
pub const PCITEST_WRITE: u32 = 2147766276;
pub const PPCLAIM: u32 = 536899723;
pub const VIDIOC_S_JPEGCOMP: u32 = 2156680766;
pub const IPMICTL_UNREGISTER_FOR_CMD: u32 = 1073899791;
pub const VHOST_SET_FEATURES: u32 = 2148052736;
pub const TOSHIBA_ACPI_SCI: u32 = 3222828177;
pub const VIDIOC_DQBUF: u32 = 3225703953;
pub const BTRFS_IOC_BALANCE_PROGRESS: u32 = 1140888610;
pub const BTRFS_IOC_SUBVOL_SETFLAGS: u32 = 2148045850;
pub const ATMLEC_MCAST: u32 = 536895954;
pub const MMTIMER_GETFREQ: u32 = 1074031874;
pub const VIDIOC_G_SELECTION: u32 = 3225441886;
pub const RTC_ALM_SET: u32 = 2149871623;
pub const PPPOEIOCSFWD: u32 = 2147791104;
pub const IPMICTL_GET_MAINTENANCE_MODE_CMD: u32 = 1074030878;
pub const FS_IOC_ENABLE_VERITY: u32 = 2155898501;
pub const NILFS_IOCTL_GET_BDESCS: u32 = 3222826631;
pub const FDFMTEND: u32 = 536871497;
pub const DMA_BUF_SET_NAME: u32 = 2147770881;
pub const UI_BEGIN_FF_UPLOAD: u32 = 3227538888;
pub const RTC_UIE_ON: u32 = 536899587;
pub const PPRELEASE: u32 = 536899724;
pub const VFIO_IOMMU_UNMAP_DMA: u32 = 536886130;
pub const VIDIOC_OMAP3ISP_PRV_CFG: u32 = 3225179842;
pub const GPIO_GET_LINEHANDLE_IOCTL: u32 = 3245126659;
pub const VFAT_IOCTL_READDIR_BOTH: u32 = 1108898305;
pub const NVME_IOCTL_ADMIN_CMD: u32 = 3225964097;
pub const VHOST_SET_VRING_KICK: u32 = 2148052768;
pub const BTRFS_IOC_SUBVOL_CREATE_V2: u32 = 2415957016;
pub const BTRFS_IOC_SNAP_CREATE: u32 = 2415956993;
pub const SONYPI_IOCGBAT2CAP: u32 = 1073903108;
pub const PPNEGOT: u32 = 2147774609;
pub const NBD_PRINT_DEBUG: u32 = 536914694;
pub const BTRFS_IOC_INO_LOOKUP_USER: u32 = 3489698878;
pub const BTRFS_IOC_GET_SUBVOL_ROOTREF: u32 = 3489698877;
pub const FS_IOC_REMOVE_ENCRYPTION_KEY_ALL_USERS: u32 = 3225445913;
pub const BTRFS_IOC_FS_INFO: u32 = 1140888607;
pub const VIDIOC_ENUM_FMT: u32 = 3225441794;
pub const VIDIOC_G_INPUT: u32 = 1074026022;
pub const VTPM_PROXY_IOC_NEW_DEV: u32 = 3222577408;
pub const DFL_FPGA_FME_ERR_GET_IRQ_NUM: u32 = 1074050691;
pub const ND_IOCTL_DIMM_FLAGS: u32 = 3221769731;
pub const BTRFS_IOC_QUOTA_RESCAN: u32 = 2151715884;
pub const MMTIMER_GETCOUNTER: u32 = 1074031881;
pub const MATROXFB_GET_OUTPUT_MODE: u32 = 3221516026;
pub const BTRFS_IOC_QUOTA_RESCAN_WAIT: u32 = 536908846;
pub const RIO_CM_CHAN_BIND: u32 = 2148033285;
pub const HIDIOCGRDESC: u32 = 1342457858;
pub const MGSL_IOCGIF: u32 = 536898827;
pub const VIDIOC_S_OUTPUT: u32 = 3221509679;
pub const HIDIOCGREPORTINFO: u32 = 3222030345;
pub const WDIOC_GETBOOTSTATUS: u32 = 1074026242;
pub const VDUSE_VQ_GET_INFO: u32 = 3224404245;
pub const ACRN_IOCTL_ASSIGN_PCIDEV: u32 = 2149884501;
pub const ACRN_IOCTL_PM_GET_CPU_STATE: u32 = 3221791328;
pub const ACRN_IOCTL_DESTROY_VM: u32 = 536912401;
pub const ACRN_IOCTL_SET_PTDEV_INTR: u32 = 2148835923;
pub const ACRN_IOCTL_CREATE_IOREQ_CLIENT: u32 = 536912434;
pub const ACRN_IOCTL_IRQFD: u32 = 2149098097;
pub const ACRN_IOCTL_CREATE_VM: u32 = 3224412688;
pub const ACRN_IOCTL_INJECT_MSI: u32 = 2148573731;
pub const ACRN_IOCTL_ATTACH_IOREQ_CLIENT: u32 = 536912435;
pub const ACRN_IOCTL_RESET_PTDEV_INTR: u32 = 2148835924;
pub const ACRN_IOCTL_NOTIFY_REQUEST_FINISH: u32 = 2148049457;
pub const ACRN_IOCTL_SET_IRQLINE: u32 = 2148049445;
pub const ACRN_IOCTL_START_VM: u32 = 536912402;
pub const ACRN_IOCTL_SET_VCPU_REGS: u32 = 2166923798;
pub const ACRN_IOCTL_SET_MEMSEG: u32 = 2149622337;
pub const ACRN_IOCTL_PAUSE_VM: u32 = 536912403;
pub const ACRN_IOCTL_CLEAR_VM_IOREQ: u32 = 536912437;
pub const ACRN_IOCTL_UNSET_MEMSEG: u32 = 2149622338;
pub const ACRN_IOCTL_IOEVENTFD: u32 = 2149622384;
pub const ACRN_IOCTL_DEASSIGN_PCIDEV: u32 = 2149884502;
pub const ACRN_IOCTL_RESET_VM: u32 = 536912405;
pub const ACRN_IOCTL_DESTROY_IOREQ_CLIENT: u32 = 536912436;
pub const ACRN_IOCTL_VM_INTR_MONITOR: u32 = 2147787300;
