use clap::{arg, command};

fn main() {
    let matches = command!() // requires `cargo` feature
        .arg(arg!([NAME]))
        .get_matches();

    println!("NAME: {:?}", matches.get_one::<String>("NAME"));
}
